"""

Analytic integral for vectorized field / potential computation

"""

__all__ = [
    "c_coeffs",
    "d_distance",
    "gamma0",
    "omega",
    "potential_dipoles",
    "potential_vertex_dipoles",
    "triangle_potential_approx",
    "triangle_potential_dipole_linear",
    "triangle_potential_uniform",
    "x_distance",
    "x_distance2",
]

import numpy as np


def determinant(a):
    """Faster determinant for the two last dimensions of 'a'"""
    det = a[..., 0, 0] * (a[..., 1, 1] * a[..., 2, 2] - a[..., 2, 1] * a[..., 1, 2])
    det += a[..., 0, 1] * (a[..., 1, 2] * a[..., 2, 0] - a[..., 2, 2] * a[..., 1, 0])
    det += a[..., 0, 2] * (a[..., 1, 0] * a[..., 2, 1] - a[..., 2, 0] * a[..., 1, 1])
    return det


def norm(vecs):
    """Faster vector norm for the last dimension of 'vecs'"""
    return np.sqrt(np.einsum("...i,...i", vecs, vecs))


def cross(r1, r2):
    """Cross product without overhead for the last dimensions of 'r1' and 'r2'"""
    result = np.zeros(r1.shape)
    result[..., 0] = r1[..., 1] * r2[..., 2] - r1[..., 2] * r2[..., 1]
    result[..., 1] = r1[..., 2] * r2[..., 0] - r1[..., 0] * r2[..., 2]
    result[..., 2] = r1[..., 0] * r2[..., 1] - r1[..., 1] * r2[..., 0]
    return result


def gamma0(R, reg=1e-13, symmetrize=True):
    """1/r integrals over the edges of a triangle called gamma_0
    (line charge potentials).

    **NOTE: MAY NOT BE VERY PRECISE FOR POINTS DIRECTLY AT TRIANGLE
    EDGES.**

    Parameters
    ----------
    R : ndarray (..., N_triverts, xyz)
        displacement vectors (r-r') between Neval evaluation points (r)
        and the 3 vertices of the Ntri triangles/triangle.
    reg: float, a small value added to the arguments of the logarithm,
         regularizes the values very close to the line segments
    symmetrize: recalculates the result for by mirroring
                the evaluation points with respect the line segment
                mid point to get rid off the badly behaving points on the
                negative extension of the line segment


    Returns
    -------
    res: array (Neval, Nverts)
        The analytic integrals for each vertex/edge

    """
    edges = np.roll(R[0], 2, -2) - np.roll(R[0], 1, -2)
    dotprods1 = np.einsum("...i,...i", np.roll(R, 1, -2), edges)
    dotprods2 = np.einsum("...i,...i", np.roll(R, 2, -2), edges)
    en = norm(edges)
    del edges
    n = norm(R)
    # Regularize s.t. neither the denominator or the numerator can be zero
    # Avoid numerical issues directly at the edge
    nn1 = np.roll(n, 1, -1) * en
    nn2 = np.roll(n, 2, -1) * en
    res = np.log((nn1 + dotprods1 + reg) / (nn2 + dotprods2 + reg))

    # Symmetrize the result since on the negative extension of the edge
    # there's division of two small values resulting numerical instabilities
    # (also incompatible with adding the reg value)
    if symmetrize:
        mask = ((np.abs(dotprods1 + nn1)) < 1e-12) * (dotprods1 + dotprods2 < 0)
        res[mask] = -np.log(
            (nn1[mask] - dotprods1[mask] + reg) / (nn2[mask] - dotprods2[mask] + reg)
        )

    res /= en
    return -res


def omega(R):
    """Calculate the solid angle of a triangles

    see
    A. Van Oosterom and J. Strackee
    IEEE TRANSACTIONS ON BIOMEDICAL ENGINEERING,
    VOL. BME-30, NO. 2, 1983

    Parameters
    ----------
    R : ndarray (Neval, (Ntri), N_triverts, xyz)
        displacement vectors (r-r') of Ntri triangles
        and Neval evaluation points for the 3 vertices
        of the triangles/triangle.

        The shape of R can any with the constraint that
        the last dimenion corrsponds to coordinates (x, y, z) and the
        second last dimension to triangle vertices (vert1, vert2, vert3)

    Returns
    -------
    sa: (Neval, (Ntri))
        Solid angles of subtened by triangles at evaluation points
    """
    # Distances
    d = norm(R)
    # Scalar triple products
    stp = determinant(R)
    # Denominator
    denom = np.prod(d, axis=-1)
    for i in range(3):
        j = (i + 1) % 3
        k = (i + 2) % 3
        #        denom += np.sum(R[..., i, :]*R[..., j, :], axis=-1)*d[..., k]
        denom += np.einsum("...i,...i,...", R[..., i, :], R[..., j, :], d[..., k])
    # Solid angles
    sa = -2 * np.arctan2(stp, denom)
    return sa


def x_distance(R, tn, ta=None):
    """Signed distances in the triangle planes from the opposite
    edge towards the node for all evaluation points in R

    The distances are normalized to one at the node if areas are given
    The distances are multiplied by the edge lenght if areass are None

    Parameters:

        R: ndarray (... Ntri, Nverts, xyz)
            displacement vectors (coordinates)
        tn: ndarray (Ntri, 3)
            triangle normals
        ta: ndarray (Ntri)
            triangle areas
            if None, normalizization with double area is not carried out

    returns:
        ndaarray (..., Ntri, N_triverts (3)), distance in the triangle plane
    """
    edges = np.roll(R[0], 2, -2) - np.roll(R[0], 1, -2)
    if ta is not None:
        edges /= 2 * ta[:, None, None]
    edges = -cross(edges, tn[:, None, :])
    return np.einsum("...k,...k->...", np.roll(R, 1, -2), edges)


def x_distance2(mesh):
    """Signed distances in the triangle planes from the opposite
    edge towards the node for all evalution points in R
    """
    # TODO: with gradient, needs mesh info
    pass


def d_distance(R, tn):
    """Signed distance from the triangle plane for each triangle

    Parameters:

        R: ndarray (... Ntri, Nverts, xyz)
            displacement vectors (coordinates)
        tn: ndarray (Ntri, 3)
            triangle normals

    Returns:

        ndarray (..., Ntri, N_triverts (3)) of signed distances
    """
    return np.einsum("...ki,ki->...k", np.take(R, 0, -2), tn)


def c_coeffs(R, ta):
    """Cotan-coeffs

    Parameters:

        R: ndarray (... Ntri, Nverts, xyz)
            displacement vectors (coordinates)
        ta: ndarray (Ntri)
            triangle areas


    Returns:

        ndarray (..., Ntri, N_triverts (3))
    """
    edges = np.roll(R[0], 2, -2) - np.roll(R[0], 1, -2)
    return np.einsum("...ik,...jk->...ij", edges, edges / (2 * ta[:, None, None]))


def triangle_potential_uniform(R, tn, planar=False):
    """1/r potential of a uniform triangle

    for original derivation see
    A. S. Ferguson, Xu Zhang and G. Stroink,
    "A complete linear discretization for calculating the magnetic field
    using the boundary element method,"
    in IEEE Transactions on Biomedical Engineering,
    vol. 41, no. 5, pp. 455-460, May 1994.
    doi: 10.1109/10.293220


    Parameters
    ----------

    R : (Neval, (Ntri), 3, 3) array
        Displacement vectors (Neval, (Ntri), Ntri_verts, xyz)
    tn : ((Ntri), 3) array
        Triangle normals (Ntri, dir)
    planar: boolean
        If True, assume all the triangles and the evaluation points
                are on the same plane (for speed), leaves out the
                omega term

    Returns
    -------
    result: result:  ndarray (Neval, (Ntri))
        Resultant 1/r potential for each triangle (Ntri)
        at the field evaluation points (Neval)

    """
    x = x_distance(R, tn, None)
    result = np.einsum("...i,...i", gamma0(R), x)
    if not planar:
        result += d_distance(R, tn) * omega(R)
    return result


def triangle_potential_approx(Rcenters, ta, reg=1e-12):
    """1/r potential of a uniform triangle using centroid approximation

    Calculates 1/R potentials for triangle centroids
    (The singularity at the centroid is handled with the very small
    reg value, but anyway the values close to the centroid are inexact)

    Parameters
    ----------
    Rcenters : (N, (Ntri), 3) array
        Displacement vectors (Neval, Ntri, xyz)
        from triangle centers
    ta : (Ntri) array
        Triangle areas

    reg: float
        Regularization value used in approximation

    Returns
    -------
    result: result:  ndarray (...., Ntri, Ntri_verts)
        Resultant 1/r potential for each node (Ntri_verts)
        in each triangle (Ntri) in the displacement vectors R

    """
    result = ta / (norm(Rcenters) + reg)
    return result


def potential_dipoles(R, face_normals, face_areas):
    """Approximate the potential of linearly varying dipole density by
        by dipoles at each face

    Parameters
            R : ndarray (Neval, Ntri, Ntri_verts, N_xyz)
                Displacement vectors
            face_normals: ndarray (Ntri, 3)
                normals for each triangle
            face_areas: ndarray (Ntri,)
                areas for each triangle

    Return
        Potential approximation for vertex in each face
        pot: (Neval, Ntri, Ntriverts)
    """
    nn = face_normals
    # Calculate quadrature points corresponding to linear shape functions (Ok?)
    weights = np.array([[0.5, 0.25, 0.25], [0.25, 0.5, 0.25], [0.25, 0.25, 0.5]])
    #    weights = np.eye(3)
    #    weights = np.ones((3,3))/3
    # Combine vertices for quadrature points
    Rquad = np.einsum("...ij,ik->...kj", R, weights)
    pot = np.einsum("ik, ...ijk->...ij", nn, Rquad) / (norm(Rquad) ** 3)
    pot = pot * (face_areas[:, None] / 3)

    return pot


def potential_vertex_dipoles(R, vertex_normals, vertex_areas):
    """Approximate the potential of linearly varying dipole density by
        by dipoles at each vertex

    Parameters
            R : ndarray (Neval, Nvertex, N_xyz)
                Displacement vectors
            vertex_normals: ndarray (Nvertex, 3)
                normals for each triangle
            vertex_areas: ndarray (Nvertex,)
                areas for each triangle

    Return
        Potential approximation for vertex in each face
        pot: (Neval, Ntri, Ntriverts)
    """
    nn = vertex_normals
    pot = np.einsum("ik, lik->li", nn, R) / (norm(R) ** 3)
    pot *= vertex_areas

    return pot


def triangle_potential_dipole_linear(R, tn, ta):
    """Potential of dipolar density with magnitude of a
    linear shape function on a triangle, "omega_i" in de Munck's paper

    for the original derivation, see:
    J. C. de Munck, "A linear discretization of the volume mesh_conductor
    boundary integral equation using analytically integrated elements
    (electrophysiology application),"
    in IEEE Transactions on Biomedical Engineering,
    vol. 39, no. 9, pp. 986-990, Sept. 1992.
    doi: 10.1109/10.256433

    Parameters
    ----------

    R : (..., Ntri, 3, 3) array
        Displacement vectors (...., Ntri, Ntri_verts, xyz)
    tn : ((Ntri), 3) array
        Triangle normals (Ntri, dir)
    ta : (Ntri), array
        Triangle areas (Ntri, dir)

    Returns
    -------
    result:  ndarray (...., Ntri, Ntri_verts)
        Resultant dipolar potential for each shape functions (Ntri_verts)
        in each triangle (Ntri) at the points
        corresponding to displacement vectors in R

    """

    result = np.einsum(
        "...i,...ij,...->...j",
        gamma0(R),
        c_coeffs(R, ta),
        d_distance(R, tn),
        optimize=True,
    )
    x_dists = x_distance(R, tn, ta)
    result -= x_dists * omega(R)[..., :, None]

    return result



"""
One-element fields
======================

"""


import numpy as np

from bfieldtools.mesh_magnetics import (
    scalar_potential_coupling,
    vector_potential_coupling,
)
from bfieldtools.mesh_magnetics import (
    magnetic_field_coupling,
    magnetic_field_coupling_analytic,
)

import trimesh
from mayavi import mlab

# Define the element
x = np.sin(np.pi / 6)
y = np.cos(np.pi / 6)
points0 = np.array(
    [[0, 0, 0], [1, 0, 0], [x, y, 0], [-x, y, 0], [-1, 0, 0], [-x, -y, 0], [x, -y, 0]]
)

tris0 = np.array([[0, 1, 2], [0, 2, 3], [0, 3, 4], [0, 4, 5], [0, 5, 6], [0, 6, 1]])
mesh = trimesh.Trimesh(points0, tris0)
scalars = np.zeros(7)
scalars[0] = 1

#%%
#%% Plot element
def plot_element():
    # Stream function
    s1 = mlab.triangular_mesh(*points0.T, tris0, scalars=scalars, colormap="viridis")
    # Stream lines
    s2 = mlab.triangular_mesh(*points0.T, tris0, scalars=scalars, colormap="viridis")
    s2.enable_contours = True
    s2.actor.mapper.scalar_range = np.array([0.0, 1.0])
    s2.actor.mapper.scalar_visibility = False
    s2.actor.property.render_lines_as_tubes = True
    s2.actor.property.line_width = 3.0


mlab.figure(bgcolor=(1, 1, 1))
plot_element()

#%%
#%% Scalar potential
points = np.array([[0.01, 1, 1], [0.01, 1, -1], [0.01, -1, -1], [0.01, -1, 1]]) * 2
tris = np.array([[0, 1, 2], [2, 3, 0]])
mesh2 = trimesh.Trimesh(points, tris)
for ii in range(7):
    mesh2 = mesh2.subdivide()

U = scalar_potential_coupling(mesh, mesh2.vertices, multiply_coeff=True) @ scalars
mlab.figure(bgcolor=(1, 1, 1))
s3 = mlab.triangular_mesh(*mesh2.vertices.T, mesh2.faces, scalars=U, colormap="bwr")
s3.enable_contours = True
s3.contour.minimum_contour = -5.2e-07
s3.contour.maximum_contour = 5.2e-07
s3.actor.property.render_lines_as_tubes = True
s3.actor.property.line_width = 3.0
s3.scene.x_plus_view()
plot_element()

#%%
#%% Vector potential
points = np.array([[1, 1, 0.01], [1, -1, 0.01], [-1, -1, 0.01], [-1, 1, 0.01]]) * 2
tris = np.array([[0, 1, 2], [2, 3, 0]])
mesh3 = trimesh.Trimesh(points, tris)
for ii in range(5):
    mesh3 = mesh3.subdivide()
A = vector_potential_coupling(mesh, mesh3.vertices) @ scalars
mlab.figure(bgcolor=(1, 1, 1))
vectors = mlab.quiver3d(*mesh3.vertices.T, *A, mode="2ddash", color=(0, 0, 1))
vectors.glyph.glyph_source.glyph_position = "center"
vectors.actor.property.render_lines_as_tubes = True
vectors.actor.property.line_width = 3.0
plot_element()

#%%
#%% Magnetic field and its magnitude
from bfieldtools.viz import plot_data_on_vertices

points = (
    np.array([[0.0, 1, 1.001], [0.0, 1, -1], [0.0, -1, -1], [0.0, -1, 1.001]]) * 1.1
)
tris = np.array([[0, 1, 2], [2, 3, 0]])
mesh2 = trimesh.Trimesh(points, tris)
for ii in range(6):
    mesh2 = mesh2.subdivide()

# B0 = magnetic_field_coupling(mesh, mesh2.vertices) @ scalars
B1 = magnetic_field_coupling_analytic(mesh, mesh2.vertices) @ scalars
plot_data_on_vertices(
    mesh2, np.linalg.norm(B1, axis=1), ncolors=32, colormap="viridis", vmax=1.5e-6
)
vectors = mlab.quiver3d(
    *mesh2.vertices.T, *B1.T, mode="arrow", color=(1, 0, 1), scale_factor=5e4
)
vectors.glyph.glyph_source.glyph_position = "center"
# vectors = mlab.quiver3d(*mesh2.vertices.T, *B0.T, mode="arrow", color=(1, 0, 0))
# vectors.glyph.glyph_source.glyph_position = "center"
plot_element()
vectors.scene.x_plus_view()



"""
Mutual inductance of wire loops
===============================

In this example, we demonstrate how to compute the mutual inductance between two sets of polyline wire loops.

"""


from bfieldtools.line_conductor import LineConductor
from bfieldtools.mesh_conductor import StreamFunction
from bfieldtools.suhtools import SuhBasis
from bfieldtools.utils import load_example_mesh

import numpy as np
from mayavi import mlab

#%%
# We create a set of wire loops by picking a single (arbitrary) surface-harmonic mode
# from a plane mesh.  Finally, we discretize the  mode into a set of wire loops, which we plot.

mesh = load_example_mesh("10x10_plane")
mesh.apply_scale(0.1)  # Downsize from 10 meters to 1 meter
N_contours = 20

sb = SuhBasis(mesh, 10)  # Construct surface-harmonics basis
sf = StreamFunction(
    sb.basis[:, 1], sb.mesh_conductor
)  # Turn single mode into a stream function
c = LineConductor(
    mesh=mesh, scalars=sf, N_contours=N_contours
)  # Discretize the stream function into wire loops


# Plot loops for testing
c.plot_loops(origin=np.array([0, -100, 0]))


#%%
# Now, we create a shifted copy of the wire loops, and the calculate the
# mutual_inductance between two sets of line conductors


mesh2 = mesh.copy()
mesh2.vertices[:, 1] += 1
c2 = LineConductor(mesh=mesh2, scalars=sf, N_contours=N_contours)
fig = c.plot_loops(origin=np.array([0, -100, 0]))
c2.plot_loops(figure=fig, origin=np.array([0, -100, 0]))

Mself = c.line_mutual_inductance(c, separate_loops=True, radius=1e-3)
M2 = c.line_mutual_inductance(c2, separate_loops=True)

#%%
# Now, we plot the inductance matrices

import matplotlib.pyplot as plt

ff, ax = plt.subplots(1, 2, figsize=(12, 8))
plt.sca(ax[0])
plt.matshow(Mself, fignum=0)
plt.title("Inductance matrix of the first set of wire loops")
plt.sca(ax[1])
plt.matshow(M2, fignum=0)
plt.title("Mutual inductance matrix between the sets of wire loops")

ff.tight_layout()

#%%
# The inductance derived from the continous current density
# ---------------------------------------------------------
# 1) Magnetic energy of a inductor is E = 0.5*L*I^2
# 2) For unit current I=1 the inductance is L=2*E
# 3) The total current of a stream function (sf) integrated over
#    the from minimum to maximum is dsf = max(sf) - min(sf)
# 4) When discretized to N conductors the current per conductor is
#    I =  dsf / N
# 5) When sf is normalized such that I=1, i.e., dsf = N
#    the inductance approximated by the continous stream function is
#    L = 2*sf.magnetic_energy

scaling = N_contours / (sf.max() - sf.min())
L_approx = 2 * sf.magnetic_energy * (scaling**2)

print("Inductance based on the continuous current density", L_approx)
print("Inductance based on r=1mm wire", np.sum(Mself))



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

import os
import os.path as op
import sys

from datetime import date

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------

project = "bfieldtools"
td = date.today()
copyright = "2019-%s, bfieldtools developers. Last updated on %s" % (
    td.year,
    td.isoformat(),
)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.


from pybtex.style.formatting.unsrt import Style as UnsrtStyle
from pybtex.style.formatting.unsrt import pages, date
from pybtex.style.formatting import toplevel
from pybtex.style.template import (
    field,
    first_of,
    href,
    join,
    names,
    optional,
    optional_field,
    sentence,
    tag,
    together,
    words,
    node,
)
from pybtex.plugin import register_plugin
from pybtex.richtext import Symbol, Text

# bibtex style
class MyStyle(UnsrtStyle):
    def __init__(self):
        super().__init__()
        self.abbreviate_names = True

    # This could be modified
    def get_article_template(self, e):
        volume_and_pages = first_of[
            # volume and pages, with optional issue number
            optional[
                join[field("volume"), optional["(", field("number"), ")"], ":", pages],
            ],
            # pages only
            words["pages", pages],
        ]
        template = toplevel[
            self.format_names("author"),
            self.format_title(e, "title"),
            sentence[tag("em")[field("journal")], optional[volume_and_pages], date],
            self.format_web_refs(e),
            sentence[optional_field("note")],
        ]
        return template


register_plugin("pybtex.style.formatting", "mystyle", MyStyle)

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.autosummary",
    "sphinx_fontawesome",
    "numpydoc",
    "sphinx.ext.intersphinx",
    "sphinx_gallery.gen_gallery",
    "sphinx_bootstrap_theme",
    "sphinxcontrib.bibtex",
    "sphinx.ext.mathjax",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# External imports
autodoc_mock_imports = [
    "numpy",
    "scipy",
    "trimesh",
    "mayavi",
    "cvxopt",
    "cvxpy",
    "numba",
    "psutil",
]

# Autodoc config

# Concatenate class docstring and __init__
autoclass_content = "both"
# Order by source order, not alphabetical
autodoc_member_order = "bysource"

# Configure sphinx-gallery

scrapers = ("matplotlib",)
try:
    from mayavi import mlab

    # Do not pop up any mayavi windows while running the
    # examples. These are very annoying since they steal the focus.
    mlab.options.offscreen = True
    # hack to initialize the Mayavi Engine
    mlab.test_plot3d()
    mlab.close()
except Exception:
    pass
else:
    scrapers += ("mayavi",)

sphinx_gallery_conf = {
    "examples_dirs": "../examples",  # path to your example scripts
    "gallery_dirs": "auto_examples",  # path where to save gallery generated examples
    "filename_pattern": ".py",  # which examples are executed for plots etc
    "image_scrapers": scrapers,
    "abort_on_example_error": False,
    "download_section_examples": False,
    "show_memory": True,
    "reference_url": {
        # The module you locally document uses None
        "sphinx_gallery": None,
    },
    "backreferences_dir": "./generated/",
    "ignore_pattern": r"/test",
}


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/devdocs", None),
    "scipy": ("https://scipy.github.io/devdocs", None),
    "matplotlib": ("https://matplotlib.org", None),
    "numba": ("https://numba.pydata.org/numba-doc/latest", None),
    "joblib": ("https://joblib.readthedocs.io/en/latest", None),
    "mayavi": ("http://docs.enthought.com/mayavi/mayavi", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
}

numpydoc_show_class_members = False

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#

html_theme = "bootstrap"

html_theme_options = {
    "bootswatch_theme": "readable",
    "navbar_sidebarrel": False,  # Render the next/prev links in navbar?
    "navbar_pagenav": False,
    "navbar_class": "navbar",
    "bootstrap_version": "3",
    "navbar_links": [
        ("Overview", "overview"),
        ("Install", "installation"),
        ("Literature", "literature"),
        ("Examples", "auto_examples/index"),
        ("API", "api_reference"),
    ],
    # HTML navbar class (Default: "navbar") to attach to <div> element.
    # For black navbar, do "navbar navbar-inverse"
    #'navbar_class': "navbar navbar-inverse",
    # Fix navigation bar to top of page?
    # Values: "true" (default) or "false"
    "navbar_fixed_top": "true",
    # Location of link to source.
    # Options are "nav" (default), "footer" or anything else to exclude.
    "source_link_position": "footer",
    "navbar_class": "navbar navbar-inverse",
    "globaltoc_depth": 1,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# A list of paths that contain extra files not directly related to
# the documentation, such as robots.txt or .htaccess.
# Relative paths are taken as relative to the configuration directory.
# They are copied to the output directory. They will overwrite any
# existing file of the same name.
html_extra_path = ["_extra"]

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/favicon.ico"


# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/logo_simple.png"
html_logo = "_static/logo_simple.png"

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = True
html_copy_source = False


# autosummary_generate = True


def append_attr_meth_examples(app, what, name, obj, options, lines):
    """Append SG examples backreferences to method and attr docstrings."""
    # NumpyDoc nicely embeds method and attribute docstrings for us, but it
    # does not respect the autodoc templates that would otherwise insert
    # the .. include:: lines, so we need to do it.
    # Eventually this could perhaps live in SG.
    if what in ("attribute", "method"):
        size = os.path.getsize(
            op.join(op.dirname(__file__), "generated", "%s.examples" % (name,))
        )
        if size > 0:
            lines += """
            .. rubric:: Examples using ``{0}``:

            .. include:: {1}.examples
            :start-line: 5

            .. raw:: html

            <div style="clear:both"></div>
            """.format(
                name.split(".")[-1], name
            ).split(
                "\n"
            )


def setup(app):
    """Set up the Sphinx app."""
    app.connect("autodoc-process-docstring", append_attr_meth_examples)



import numpy as np
import matplotlib.pyplot as plt
import trimesh
from mayavi import mlab
from bfieldtools.thermal_noise import (
    compute_current_modes,
    noise_covar,
    noise_var,
    visualize_current_modes,
)
from bfieldtools.mesh_magnetics import magnetic_field_coupling

import pkg_resources


font = {"family": "normal", "weight": "normal", "size": 16}
plt.rc("font", **font)

# Fix the simulation parameters
d = 1e-3
sigma = 3.8e7
T = 293
kB = 1.38064852e-23
mu0 = 4 * np.pi * 1e-7
freqs = np.array((0,))


Nchunks = 8
quad_degree = 2

##############################################################################
# Unit sphere
# ------------


Np = 10
radius = np.linspace(0.1, 1, Np)
fp = np.zeros((1, 3))

B = np.zeros((Np, 3))
for i in range(Np):
    mesh = trimesh.load(
        pkg_resources.resource_filename("bfieldtools", "example_meshes/unit_sphere.stl")
    )
    mesh.apply_scale(radius[i])

    B_coupling = magnetic_field_coupling(mesh, fp, analytic=True)

    vl = compute_current_modes(
        obj=mesh,
        T=T,
        resistivity=1 / sigma,
        thickness=d,
        mode="AC",
        return_eigenvals=False,
    )

    #    scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5),
    #               size=(800, 800))
    #    visualize_current_modes(mesh,vl[:,:,0], 8, 1)

    vl[:, 0] = np.zeros(vl[:, 0].shape)  # fix DC-component

    Btemp = noise_var(B_coupling, vl)

    B[i] = Btemp[:, :, 0]

scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
s = mlab.triangular_mesh(*mesh.vertices.T, mesh.faces)
scene.scene.z_minus_view()
surface = scene.children[0].children[0].children[0].children[0]
surface.actor.property.representation = "wireframe"
surface.actor.mapper.scalar_visibility = False
scene.scene.camera.position = [0.0, 0.0, -5.530686305704514]
scene.scene.camera.focal_point = [0.0, 0.0, 0.0]
scene.scene.camera.view_angle = 30.0
scene.scene.camera.view_up = [0.0, 1.0, 0.0]
scene.scene.camera.clipping_range = [3.485379442647469, 8.118646600290083]
scene.scene.camera.compute_view_plane_normal()
scene.scene.render()
scene.scene.camera.position = [0.0, 0.0, -4.570815128681416]
scene.scene.camera.focal_point = [0.0, 0.0, 0.0]
scene.scene.camera.view_angle = 30.0
scene.scene.camera.view_up = [0.0, 1.0, 0.0]
scene.scene.camera.clipping_range = [2.535106977394602, 7.1443773556116374]
scene.scene.camera.compute_view_plane_normal()
scene.scene.render()


Ban = mu0 * np.sqrt(2 * sigma * d * kB * T / (3 * np.pi * (radius) ** 2))

plt.figure(figsize=(5, 5))
plt.semilogy(radius, Ban * 1e15, linewidth=2, label="Analytic")
plt.semilogy(
    radius,
    np.sqrt(B[:, 2]) * 1e15,
    "x",
    markersize=10,
    markeredgewidth=2,
    label="Numerical",
)
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False)
plt.xlabel("R (m)")
plt.ylabel(r"$B_z$ noise at DC (fT/rHz)")
plt.ylim(5, 100)
plt.tight_layout()


RE = np.abs((np.sqrt(B[:, 2]) - Ban)) / np.abs(Ban) * 100
plt.figure()
plt.plot(radius, RE)
plt.xlabel("Sphere radius")
plt.ylabel("Relative error (%)")



"""
Design a field enclosed in a volume
================================================
"""

import numpy as np
import matplotlib.pyplot as plt
from mayavi import mlab

# import trimesh

from bfieldtools.mesh_conductor import MeshConductor, StreamFunction
from bfieldtools.mesh_magnetics import magnetic_field_coupling as compute_C
from bfieldtools.mesh_magnetics import (
    magnetic_field_coupling_analytic as compute_C_analytic,
)
from bfieldtools.mesh_magnetics import scalar_potential_coupling as compute_U
from bfieldtools.mesh_impedance import mutual_inductance_matrix
from bfieldtools.contour import scalar_contour
from bfieldtools import sphtools
from bfieldtools.utils import load_example_mesh
from bfieldtools.mesh_calculus import mass_matrix

# domain = 'sphere'
# domain = 'cube'
domain = "combined"

if domain == "sphere":
    from trimesh.creation import icosphere

    mesh1 = icosphere(3, 0.65)
    mesh2 = icosphere(3, 0.8)
elif domain == "cube":
    from trimesh.creation import box
    from trimesh.smoothing import filter_laplacian

    mesh1 = box((1.0, 1.0, 1.0))
    mesh2 = box((1.5, 1.5, 1.5))

    for i in range(4):
        mesh1 = mesh1.subdivide()
        mesh2 = mesh2.subdivide()

    mesh1 = filter_laplacian(mesh1)
    mesh2 = filter_laplacian(mesh2, 0.9)
elif domain == "combined":
    from trimesh.creation import icosphere

    mesh1 = icosphere(4, 0.65)
    mesh2 = load_example_mesh("cube_fillet")
    mesh2.vertices -= mesh2.vertices.mean(axis=0)
    mesh2.vertices *= 0.15
#    mesh2 = mesh2.subdivide()

coil1 = MeshConductor(
    mesh_obj=mesh1,
    N_sph=3,
    inductance_nchunks=100,
    fix_normals=False,
    inductance_quad_degree=2,
)
coil2 = MeshConductor(
    mesh_obj=mesh2,
    N_sph=3,
    inductance_nchunks=100,
    fix_normals=False,
    inductance_quad_degree=2,
)

M11 = coil1.inductance
M22 = coil2.inductance
# Add rank-one matrix, so that M22 can be inverted
M22 += np.ones_like(M22) / M22.shape[0] * np.mean(np.diag(M22))
M11 += np.ones_like(M11) / M11.shape[0] * np.mean(np.diag(M11))
M21 = mutual_inductance_matrix(mesh2, mesh1)

M = np.block([[M11, M21.T], [M21, M22]])


x = y = np.linspace(-0.85, 0.85, 100)
X, Y = np.meshgrid(x, y, indexing="ij")
points = np.zeros((X.flatten().shape[0], 3))
points[:, 0] = X.flatten()
points[:, 1] = Y.flatten()

CB1 = compute_C_analytic(mesh1, points)
CB2 = compute_C_analytic(mesh2, points)

CU1 = compute_U(mesh1, points)
CU2 = compute_U(mesh2, points)

#%%
# suh = SuhBasis(mesh1, 100)

#%% Specify (spherical harmonic) field and calculate corresponding shielded field
# b10 = mesh1.vertex_normals[:, 0]
# b20 = (
#     mesh1.vertex_normals[:, 0] * mesh1.vertices[:, 0]
#     - mesh1.vertex_normals[:, 1] * mesh1.vertices[:, 1]
# )
N = mass_matrix(mesh1, lumped=True)
af, bf = sphtools.basis_fields(mesh1.vertices, 2)
b1 = N @ np.sum(bf[:, :, 2] * mesh1.vertex_normals, axis=1)
b2 = N @ np.sum(bf[:, :, 3] * mesh1.vertex_normals, axis=1)


def plot_plane(opacity=0.8):
    mlab.triangular_mesh(
        np.array([x[0], x[-1], x[-1], x[0]]),
        np.array([x[0], x[0], x[-1], x[-1]]),
        np.zeros(4),
        np.array([[0, 1, 2], [2, 3, 0]]),
        color=(0.7, 0.7, 0.7),
        opacity=opacity,
    )


for bi in (b1, b2):

    bb = np.zeros(M.shape[1])
    bb[: M11.shape[1]] = bi
    I = np.linalg.solve(M, bb)
    I1 = I[: M11.shape[1]]
    I2 = I[M11.shape[1] :]

    B1 = CB1 @ I1
    B2 = CB2 @ I2

    U1 = CU1 @ I1
    U2 = CU2 @ I2

    # Plot
    # Extract the cross-sections of the plane and the surfaces
    cc1 = scalar_contour(mesh1, mesh1.vertices[:, 2], contours=[-0.001])[0]
    cc2 = scalar_contour(mesh2, mesh2.vertices[:, 2], contours=[-0.001])[0]

    B = (B1.T + B2.T)[:2].reshape(2, x.shape[0], y.shape[0])
    lw = np.sqrt(B[0] ** 2 + B[1] ** 2)
    lw = 2 * lw / np.max(lw)
    xx = np.linspace(-1, 1, 16)
    # seed_points = 0.51*np.array([xx, -np.sqrt(1-xx**2)])
    # seed_points = np.hstack([seed_points, (0.51*np.array([xx, np.sqrt(1-xx**2)]))])
    seed_points = np.array([cc1[:, 0], cc1[:, 1]]) * 1.01
    # plt.streamplot(x,y, B[1], B[0], density=2, linewidth=lw, color='k',
    #               start_points=seed_points.T, integration_direction='both')
    U = (U1 + U2).reshape(x.shape[0], y.shape[0])
    U /= np.max(U)
    plt.figure()
    plt.contourf(X, Y, U.T, cmap="seismic", levels=40)
    # plt.imshow(U, vmin=-1.0, vmax=1.0, cmap='seismic', interpolation='bicubic',
    #           extent=(x.min(), x.max(), y.min(), y.max()))
    plt.streamplot(
        x,
        y,
        B[1],
        B[0],
        density=2,
        linewidth=lw,
        color="k",
        start_points=seed_points.T,
        integration_direction="both",
    )

    plt.plot(cc1[:, 1], cc1[:, 0], linewidth=3.0, color="gray")
    plt.plot(cc2[:, 1], cc2[:, 0], linewidth=3.0, color="gray")

    plt.xticks([])
    plt.yticks([])

    plt.axis("image")

    # Plot "coils"
    mlab.figure(bgcolor=(1, 1, 1))
    contours1 = scalar_contour(mesh1, I1, 12)
    contours2 = scalar_contour(mesh2, I2, 12)

    #    fig = plot_3d_current_loops(contours1, tube_radius=0.005, colors=(1,1,1))
    surf = mlab.triangular_mesh(
        *mesh1.vertices.T, mesh1.faces, scalars=I1, colormap="BrBG"
    )
    surf.actor.mapper.interpolate_scalars_before_mapping = True
    surf.module_manager.scalar_lut_manager.number_of_colors = 16

    #    plot_3d_current_loops(contours2, tube_radius=0.005, figure=fig, colors=(0,0,0))
    faces2_masked = mesh2.faces[
        np.linalg.norm(mesh2.triangles_center - np.array([0.75, 0.75, 0.75]), axis=1)
        > 1.2
    ]
    surf = mlab.triangular_mesh(
        *(mesh2.vertices * 0.99).T,
        faces2_masked,
        scalars=I2,
        colormap="BrBG",
        opacity=1.0
    )
    surf.actor.mapper.interpolate_scalars_before_mapping = True
    surf.module_manager.scalar_lut_manager.number_of_colors = 16
    plot_plane(0.5)


#%% Plot meshes

mlab.figure(bgcolor=(1, 1, 1))
s1 = mlab.triangular_mesh(
    *mesh1.vertices.T, mesh1.faces[:, ::-1], color=(0.5, 0.5, 0.5), opacity=1.0
)
s1.actor.property.backface_culling = True
w1 = mlab.triangular_mesh(
    *(mesh1.vertices.T + 0.009 * mesh1.vertex_normals.T),
    mesh1.faces,
    color=(
        1,
        0,
        0,
    ),
    representation="wireframe"
)
w1.actor.property.render_lines_as_tubes = True

s2 = mlab.triangular_mesh(
    *mesh2.vertices.T, mesh2.faces[:, ::-1], color=(0.5, 0.5, 0.5), opacity=1.0
)
s2.actor.property.backface_culling = True
faces2_masked = mesh2.faces[(mesh2.triangles_center @ np.array([1, 1, 1])) < 1]
w2 = mlab.triangular_mesh(
    *(mesh2.vertices.T + +0.009 * mesh2.vertex_normals.T),
    faces2_masked,
    representation="wireframe",
    color=(0, 0, 1)
)
w2.actor.property.render_lines_as_tubes = True
plot_plane()



# -*- coding: utf-8 -*-
"""
Created on Sun May  9 16:20:16 2021

@author: antti
"""

import sys

sys.path.insert(0, "C:/Users/antti/bfieldtools")
from bfieldtools import contour
import numpy as np
import matplotlib.pyplot as plt


def setup_contour_input():
    """Load example mesh and create scalars data"""
    from bfieldtools.utils import load_example_mesh

    mesh = load_example_mesh("unit_disc")

    r = np.linalg.norm(mesh.vertices, axis=1)
    scalars = (1 - r) ** 2
    scalars *= mesh.vertices[:, 0]

    return mesh, scalars


def test_contour_basis():
    mesh, scalars = setup_contour_input()
    N = 10
    polys, vals = contour.scalar_contour(
        mesh, scalars, N_contours=N, return_values=True
    )
    fourier_basis_funcs, freqs = get_fourier_basis(4)

    for polyline in polys:
        # polyline = polys[-1]
        W = fit_fourier_basis(polyline, fourier_basis_funcs, freqs, 1e-4)

        x_new = np.linspace(0, 2 * np.pi, 100)
        polyline_new = (W.T @ fourier_basis_funcs(x_new)).T
        plt.plot(polyline_new[:, 0], polyline_new[:, 1])

    for p in polys:
        plt.plot(p[:, 0], p[:, 1], "--*")

    plt.axis("equal")


def get_fourier_basis(Nfreqs):
    freqs = np.zeros(2 * Nfreqs + 1)
    for k in range(1, Nfreqs + 1):
        freqs[2 * k - 1] = k
        freqs[2 * k] = k

    def basis_funcs(x):
        M = np.zeros((2 * Nfreqs + 1, len(x)))
        M[0] = 1
        for k in range(1, Nfreqs + 1):
            M[2 * k - 1] = np.cos(k * x)
            M[2 * k] = np.sin(k * x)
        return M

    return basis_funcs, freqs


def contour_basis_matrix(polyline, basis_funcs):
    edges_lens = np.linalg.norm(np.roll(polyline, -1, 0) - polyline, axis=-1)
    x = np.cumsum(edges_lens) / np.sum(edges_lens)
    x = np.roll(x, 1, 0) * 2 * np.pi
    x[0] = 0

    return basis_funcs(x)


def fit_fourier_basis(polyline, basis_funcs, freqs, _lambda=0.01):
    M = contour_basis_matrix(polyline, basis_funcs)
    DD = np.diag(freqs**4)
    MM = M @ M.T
    m0 = np.linalg.eigvalsh(MM)[-1]

    scaling = _lambda * m0 * 4
    W = np.linalg.solve(MM + scaling * DD, M @ polyline)

    return W


if __name__ == "__main__":
    test_contour_basis()



"""
Contains functions for computing the resistance and inductance matrices of triangle surface meshes.
Includes both self- and mutual-inductance.
"""

__all__ = [
    "mesh2line_mutual_inductance",
    "mutual_inductance_matrix",
    "resistance_matrix",
    "self_inductance_matrix",
    "triangle_self_coupling",
]

from psutil import virtual_memory
import numpy as np

from .quadratures import get_quad_points, get_line_quad_points
from .mesh_magnetics import vector_potential_coupling
from .mesh_calculus import gradient_matrix, laplacian_matrix


def resistance_matrix(mesh, sheet_resistance):
    """Resistance matrix of a mesh.

    Parameters
    ----------
    mesh: Trimesh mesh object
    sheet_resistance: (N_faces) array or scalar
        "1/(sigma*d)", constant resistance for each face (or all faces if scalar)

    Returns
    -------
    R: (Nvertices x Nvertices) array
        resistance matrix of `mesh`
    """
    return -laplacian_matrix(mesh, sheet_resistance)


def self_inductance_matrix(
    mesh,
    Nchunks=None,
    quad_degree=2,
    approx_far=True,
    margin=2,
    chunk_clusters=False,
    planar=False,
    analytic_self_coupling=True,
):
    """Calculate a self inductance matrix for hat basis functions
    (stream functions) in a mesh.

    Self-coupling terms are calculated anaytically.

    Parameters
    ----------
    mesh: Trimesh mesh object
    Nchunks: int
        Number of serial chunks to divide the computation into
    quad_degree: int >= 1
        Quadrature degree (Dunavant scheme) to use. Self-inductance requires higher degree than mutual inductance
    approx_far: Boolean (True)
        If True, use approximate calculation for triangles that
        far from the source triangles using a simple quadrature
        (see integrals.triangle_potential_approx)
    margin: float
        Cut-off distance for "far" points measured in mean triangle side length
    planar: boolean
        This option is propagated to _triangle_coupling
        for planar meshes the analytic calculations can be made faster
    analytic_self_coupling: boolean
        If True: the diagonal elements obtained from _triangle_coupling are
        replaced with an analytic calculation
    Returns
    -------
    M: (Nvertices x Nvertices) array
        Self.inductance matrix of `mesh`
    """
    from .mesh_magnetics import _triangle_coupling

    if quad_degree <= 2:
        print(
            "Computing self-inductance matrix using rough quadrature (degree=%d).\
              For higher accuracy, set quad_degree to 4 or more."
            % quad_degree
        )

    # Calculate quadrature points
    weights, quadpoints = get_quad_points(
        mesh.vertices, mesh.faces, ("dunavant", quad_degree)
    )

    if Nchunks is None:
        Nchunks = _estimate_nchunks(mesh, mesh, approx_far)

    Nw = len(weights)
    Nt = len(mesh.faces)
    Nv = len(mesh.vertices)
    C = _triangle_coupling(
        mesh,
        quadpoints.reshape(-1, 3),
        Nchunks,
        approx_far,
        margin,
        chunk_clusters,
        planar,
    ).reshape(Nt, Nw, Nt)

    # Integrate over the triangles
    C = np.sum(C * weights[None, :, None], axis=1)
    C *= mesh.area_faces[:, None]
    # Symmetrize
    C = 0.5 * (C + C.T)
    if analytic_self_coupling:
        # Replace diagonal with the analytic version
        C[np.diag_indices(C.shape[0])] = triangle_self_coupling(mesh)

    # Rotated gradients (currents)
    Gx, Gy, Gz = gradient_matrix(mesh, rotated=True)
    # Vector potentials integrated over triangles
    A = (C @ Gx, C @ Gy, C @ Gz)
    # Dot product with current patterns and sum over triangle neighbourhoods
    M = A[0].T @ Gx + A[1].T @ Gy + A[2].T @ Gz

    coeff = 1e-7  # mu_0/(4*pi)
    return coeff * M


def mutual_inductance_matrix(
    mesh1, mesh2, Nchunks=None, quad_degree=1, approx_far=True, margin=2
):
    """Calculate a mutual inductance matrix for hat basis functions
    (stream functions) between two surface meshes.

    Parameters
    ----------

    mesh1: Trimesh mesh object for mesh 1
    mesh2: Trimesh mesh object for mesh 2
    Nchunks: int
        Number of serial chunks to divide the computation into
    quad_degree: int >= 1
        Quadrature degree (Dunavant scheme) to use. Self-inductance requires higher degree than mutual inductance
    approx_far: Boolean (True)
        If True, use approximate calculation for triangles that
        far from the source triangles using a simple quadrature
        (see integrals.triangle_potential_approx)
    margin: float
        Cut-off distance for "far" points measured in mean triangle side length

    Returns
    -------
    M: (Nvertices1 x Nvertices2) array
        Mutual inductance matrix between mesh1 and mesh2

    """

    if Nchunks is None:
        Nchunks = _estimate_nchunks(mesh1, mesh2, approx_far)

    # Calculate quadrature points
    # Nt x Nquad x  3 (x,y,z)
    weights, quadpoints = get_quad_points(
        mesh2.vertices, mesh2.faces, ("dunavant", quad_degree)
    )

    # Compute vector potential at quadrature points
    Nw = len(weights)
    Nt = len(mesh2.faces)
    Nv = len(mesh1.vertices)

    A = vector_potential_coupling(
        mesh1,
        quadpoints.reshape(-1, 3),
        Nchunks=Nchunks,
        approx_far=approx_far,
        margin=margin,
    ).reshape(Nt, Nw, 3, Nv)

    # Integrate over the triangles (current patterns are constant over triangles)
    A = np.sum(A * weights[None, :, None, None], axis=1)
    A *= mesh2.area_faces[:, None, None]

    Gx, Gy, Gz = gradient_matrix(mesh2, rotated=True)
    # Dot product with current patterns and sum over triangle neighbourhoods
    M = A[:, 0].T @ Gx + A[:, 1].T @ Gy + A[:, 2].T @ Gz

    return M


def _estimate_nchunks(mesh1, mesh2, approx_far):
    """Estimate the number of chunks for inductance calculations
    based on available memory
    """
    # Available RAM in megabytes
    mem = virtual_memory().available >> 20

    # Estimate of memory usage in megabytes for a single chunk, when quad_degree=2 (very close with quad_degree=1)
    mem_use = 0.033 * (len(mesh1.vertices) * len(mesh2.vertices)) ** 0.86

    print(
        "Estimating %d MiB required for %d by %d vertices..."
        % (mem_use, len(mesh1.vertices), len(mesh2.vertices))
    )

    # Chunk computation so that available memory is sufficient
    Nchunks = int(np.ceil(mem_use / mem))

    if approx_far:
        Nchunks *= 20
        print(
            "Computing inductance matrix in %d chunks (%d MiB memory free),\
              when approx_far=True using more chunks is faster..."
            % (Nchunks, mem)
        )
    else:
        print(
            "Computing inductance matrix in %d chunks since %d MiB memory is available..."
            % (Nchunks, mem)
        )

    return Nchunks


def triangle_self_coupling(mesh):
    """
    Self-coupling integrated analytically.

    Re-implemented based on some simplifications in the calculation presented in

    Poole, M.S., 2007. Improved equipment and techniques for dynamic
    shimming in high field MRI
    (Doctoral dissertation, University of Nottingham.). page 72.
    https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=475946
    """
    from .integrals import norm

    tri_points = mesh.vertices[mesh.faces]
    # e_i, e_j, e_k
    edges = np.roll(tri_points, 1, -2) - np.roll(tri_points, 2, -2)
    # e_j . e_k, e_k . e_i, e_i . e_j
    dotprods = np.einsum("...i,...i", np.roll(edges, 2, -2), np.roll(edges, 1, -2))
    enorms = norm(edges)

    integrals = []
    for i in range(3):
        j = (i + 1) % 3
        k = (i + 2) % 3
        a1 = enorms[:, i] * enorms[:, j]
        a2 = enorms[:, i] * enorms[:, k]
        d1 = dotprods[:, k]
        d2 = dotprods[:, j]
        logterm = np.log((a1 - d1) * (a2 - d2) / ((a1 + d1) * (a2 + d2)))
        integrals.append(logterm / enorms[:, i])

    integral = ((4 / 6) * mesh.area_faces**2) * np.sum(integrals, axis=0)

    return integral


def mesh2line_mutual_inductance(mesh, line_vertices, quad_degree=2):
    """
    Mutual inductance of a closed line segment loop (last segment connecting to first)
    and a triangle mesh.

    Parameters
    ----------
    mesh: Trimesh mesh object
    line_vertices: points connected in index order (N_points, 3)

    Returns
    -------

    M: mutual inductance vector with shape (N_vertices,)


    """

    # Calculate quadrature points
    weights, quadpoints = get_line_quad_points(
        line_vertices, ("gauss_legendre", quad_degree)
    )
    # Ne x Nquad x  3 (x,y,z)

    segments = np.roll(line_vertices, shift=-1, axis=0) - line_vertices

    # Compute vector potential to quadrature points
    Nw = len(weights)
    Nt = len(line_vertices)
    Nv = len(mesh.vertices)
    M = vector_potential_coupling(mesh, quadpoints.reshape(-1, 3)).reshape(
        Nt, Nw, 3, Nv
    )

    # Integrate over quadrature points
    M = np.sum(M * weights[None, :, None, None], axis=2)

    # Scale by segment lengths, integrate over xyz-axis and segments
    M = np.sum(segments[:, :, None] * M, axis=(0, 1))

    return M



from bfieldtools import contour

import pytest
import numpy as np

from numpy.testing import (
    assert_array_almost_equal,
    assert_array_equal,
    assert_allclose,
    assert_equal,
)


def setup_contour_input():
    """Load example mesh and create scalars data"""
    from bfieldtools.utils import load_example_mesh

    mesh = load_example_mesh("unit_disc")

    r = np.linalg.norm(mesh.vertices, axis=1)
    scalars = r**2

    return mesh, scalars


def compare_contour_direction_to_rotated_gradient(mesh, scalars, polyline):
    """Check inner product between the polyline edges
        the rotated gradient vectors closes to the initial points of those
        edges. These should point to the same direction.


    Parameters
    ----------
    mesh : trimesh
        mesh
    scalars : ndarray
        stream function
    polyline : ndarray (N, 3)
        coordinates of points representing a polyline

    """
    from bfieldtools.mesh_calculus import gradient

    edges = polyline[1:] - polyline[:-1]
    g = gradient(scalars, mesh, rotated=True).T
    fc = mesh.vertices[mesh.faces].mean(axis=1)

    norm = np.linalg.norm
    p = polyline
    # Find closest face centers to polyline nodes
    f_inds = np.argmin(norm(p[:, None, :] - fc[None, :, :], axis=-1), axis=1)

    g_poly = g[f_inds]
    assert np.all(np.sum(g_poly[:-1] * edges, axis=1) > 0)


def compare_magnetic_field_directions(mesh, scalars, polys, test_point):
    """Check direction of magnetic calculated from mesh and stream function
        versus the direction of the field generated by contours


    Parameters
    ----------
    mesh : trimesh
        mesh
    scalars : ndarray
        stream function
    polys : list of ndarray (N, 3)
        coordinates of points representing a many polylines
    test_point: ndarray(1, 3)
        coordinates of a test point

    """
    from bfieldtools.line_magnetics import magnetic_field
    from bfieldtools.mesh_conductor import magnetic_field_coupling

    B_mesh = magnetic_field_coupling(mesh, test_point)
    b_stream_func = B_mesh @ scalars
    b_polys = np.array([magnetic_field(p, test_point) for p in polys])
    b_poly_sum = b_polys.sum(axis=0)

    assert np.sum(b_poly_sum * b_stream_func) > 0


def test_scalar_contour_direction():
    """Test the direction of scalar_contour and the field direction
    generated by the scalar contour
    """
    mesh, scalars = setup_contour_input()
    N = 10

    polys, vals = contour.scalar_contour(
        mesh, scalars, N_contours=N, return_values=True
    )

    compare_contour_direction_to_rotated_gradient(mesh, scalars, polys[-1])

    test_point = np.array([[0, 0, 1]])
    compare_magnetic_field_directions(mesh, scalars, polys, test_point)


def test_simplify_contour():

    assert True


if __name__ == "__main__":
    test_scalar_contour_direction()



"""
Examples of thermal noise computation
=====================================

Different examples:
- unit_disc: DC Bnoise of a unit disc at z-axis and comparison to analytical formula
- unit_sphere: DC Bnoise of a spherical shell at origin and comparison to analytical formula
- cylinder: DC Bnoise inside a cylindrical conductor
- AC: AC Bnoise of a unit disc at one position

Analytic formulas are from Lee and Romalis (2008)
"""


import numpy as np
import matplotlib.pyplot as plt
import trimesh
from mayavi import mlab

from bfieldtools.mesh_impedance import self_inductance_matrix, resistance_matrix
from bfieldtools.thermal_noise import (
    compute_current_modes,
    noise_covar,
    noise_var,
    visualize_current_modes,
)
from bfieldtools.mesh_magnetics import magnetic_field_coupling

import pkg_resources


font = {"family": "normal", "weight": "normal", "size": 16}
plt.rc("font", **font)

# Fix the simulation parameters
d = 100e-6  # thickness
sigma = 3.7e7  # conductivity
res = 1 / sigma  # resistivity
T = 300  # temperature
kB = 1.38064852e-23  # Boltz
mu0 = 4 * np.pi * 1e-7  # permeability of freespace
# freqs = np.array((0,))


# Nchunks = 8
# quad_degree = 2


#%%
# DC magnetic noise from unit disc

mesh = trimesh.load(
    pkg_resources.resource_filename("bfieldtools", "example_meshes/unit_disc.stl")
)
mesh.vertices, mesh.faces = trimesh.remesh.subdivide(mesh.vertices, mesh.faces)
mesh.vertices, mesh.faces = trimesh.remesh.subdivide(mesh.vertices, mesh.faces)

# Compute the AC-current modes and visualize them
vl, u = compute_current_modes(
    obj=mesh, T=T, resistivity=res, thickness=d, mode="AC", return_eigenvals=True
)

scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
visualize_current_modes(mesh, vl[:, :, 0], 42, 5, contours=True)

# Define field points on z axis
Np = 30
z = np.linspace(0.1, 1, Np)
fp = np.array((np.zeros(z.shape), np.zeros(z.shape), z)).T

B_coupling = magnetic_field_coupling(mesh, fp, analytic=True)  # field coupling matrix

# Compute noise variance
B = np.sqrt(noise_var(B_coupling, vl))

# Calculate Bz noise using analytical formula and plot the results
r = 1
Ban = (
    mu0
    * np.sqrt(sigma * d * kB * T / (8 * np.pi * z**2))
    * (1 / (1 + z**2 / r**2))
)

plt.figure()
plt.subplot(2, 1, 1)
plt.semilogy(z, Ban, label="Analytic")
plt.semilogy(z, B[:, 2, 0], "x", label="Numerical")
plt.legend(frameon=False)
plt.xlabel("Distance d/R")
plt.ylabel("DC noise Bz (T/rHz)")

plt.subplot(2, 1, 2)
plt.plot(z, np.abs((B[:, 2, 0] - Ban)) / np.abs(Ban) * 100)
plt.xlabel("Distance d/R")
plt.ylabel("Relative error (%)")
plt.tight_layout()

# Next, we compute the DC noise without reference to the inductance
vl_dc, u_dc = compute_current_modes(
    obj=mesh, T=T, resistivity=res, thickness=d, mode="DC", return_eigenvals=True
)

# Compute noise variance
B_dc = np.sqrt(noise_var(B_coupling, vl_dc))

# Compare results computed using AC and DC formulation
plt.figure()
plt.semilogy(z, B_dc[:, 2], "o", label="DC mode")
plt.semilogy(z, B[:, 2, 0], "x", label="AC mode")
plt.legend(frameon=False)
plt.xlabel("Distance d/R")
plt.ylabel("DC noise Bz (T/rHz)")
plt.tight_layout()

#%%
# DC magnetic noise in the center of sphere with different radii
# ------------

Np = 10
radius = np.linspace(0.1, 1, Np)
fp = np.zeros((1, 3))  # calculate are at origin

B = np.zeros((Np, 3))
for i in range(Np):
    mesh = trimesh.load(
        pkg_resources.resource_filename("bfieldtools", "example_meshes/unit_sphere.stl")
    )
    mesh.apply_scale(radius[i])

    B_coupling = magnetic_field_coupling(mesh, fp, analytic=True)

    vl = compute_current_modes(obj=mesh, T=T, resistivity=res, thickness=d, mode="DC")
    Btemp = noise_var(B_coupling, vl)
    B[i] = Btemp

# Analytic formula
Ban = mu0 * np.sqrt(2 * sigma * d * kB * T / (3 * np.pi * (radius) ** 2))

plt.figure(figsize=(5, 5))
plt.semilogy(radius, Ban * 1e15, linewidth=2, label="Analytic")
plt.semilogy(
    radius,
    np.sqrt(B[:, 2]) * 1e15,
    "x",
    markersize=10,
    markeredgewidth=2,
    label="Numerical",
)
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False)
plt.xlabel("Sphere radius")
plt.ylabel(r"$B_z$ noise at DC (fT/rHz)")
plt.tight_layout()


#%%
# Closed cylinder, DC noise
# --------------------------

mesh = trimesh.load(
    pkg_resources.resource_filename("bfieldtools", "example_meshes/closed_cylinder.stl")
)
mesh.vertices, mesh.faces = trimesh.remesh.subdivide(mesh.vertices, mesh.faces)

# Compute noise current modes at DC
vl = compute_current_modes(obj=mesh, T=T, resistivity=res, thickness=d, mode="DC")

# Visualize the current modes
scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
visualize_current_modes(mesh, vl, 8, 1)

# Calculate field noise along long axis of the cylinder
Np = 30
x = np.linspace(-0.95, 0.95, Np)
fp = np.array((x, np.zeros(x.shape), np.zeros(x.shape))).T

B_coupling = magnetic_field_coupling(mesh, fp, analytic=True)
B = noise_var(B_coupling, vl)

# Analytic formula valid only at the center of cylinder
a = 0.5
L = 2
rat = L / (2 * a)
Gfact = (
    1
    / (8 * np.pi)
    * (
        (3 * rat**5 + 5 * rat**3 + 2) / (rat**2 * (1 + rat**2) ** 2)
        + 3 * np.arctan(rat)
    )
)
Ban = np.sqrt(Gfact) * mu0 * np.sqrt(kB * T * sigma * d) / a


plt.figure(figsize=(5, 5))
plt.plot(x, Ban * np.ones(x.shape) * 1e15, label="Analytic", linewidth=2)
plt.plot(
    x,
    np.sqrt(B[:, 0]) * 1e15,
    "x",
    label="Numerical",
    markersize=10,
    markeredgewidth=2,
)
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False)
plt.xlabel("Distance along long axis")
plt.ylabel("DC noise along axis (fT/rHz)")
plt.tight_layout()


#%%
# Unit disc, AC noise
# ------------------

mesh = trimesh.load(
    pkg_resources.resource_filename(
        "bfieldtools", "example_meshes/unitdisc_extremelyfine.stl"
    )
)

Nfreqs = 10
freqs = np.logspace(0, 4, 15)  # freqs from 1 to 10 kHz

vl = compute_current_modes(
    obj=mesh,
    T=T,
    resistivity=res,
    thickness=d,
    mode="AC",
    freqs=freqs,
    return_eigenvals=False,
)


Np = 10
z = np.linspace(0.05, 1, Np)
fp = np.array((np.zeros(z.shape), np.zeros(z.shape), z)).T

B_coupling = magnetic_field_coupling(mesh, fp, analytic=True)

Bf = np.sqrt(noise_var(B_coupling, vl))  # noise variance

# Plot Bz noise as a function of frequency
plt.figure()
plt.loglog(freqs, Bf[:, 2, :].T * 1e15, linewidth=2)
plt.grid()
# plt.ylim(1, 20)
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False)
plt.xlabel("Frequency (Hz)")
plt.ylabel(r"$B_z$ noise (fT/rHz)")
plt.tight_layout()



"""
Contains functions for calculating the coupling of surface current density in a
triangle mesh to magnetic field as well as scalar and vector potentials.

"""

__all__ = [
    "magnetic_field_coupling",
    "magnetic_field_coupling_analytic",
    "scalar_potential_coupling",
    "vector_potential_coupling",
]

import time
import numpy as np


from .quadratures import get_quad_points
from .mesh_calculus import gradient_matrix, mass_matrix
from .integrals import triangle_potential_dipole_linear, triangle_potential_uniform
from .integrals import (
    triangle_potential_approx,
    potential_vertex_dipoles,
)


def magnetic_field_coupling(mesh, r, Nchunks=None, quad_degree=1, analytic=False):
    """
    Computes the coupling matrix which gives the magnetic field at
    target points due to currents (stream function) on a surface mesh.

    Parameters
    ----------

    mesh: Trimesh mesh object
        mesh describing the geometry of the field source
    r: ndarray (Np, 3)
        evalution points
    quad_degree: int >= 1
        Quadrature degree (Dunavant scheme) to use.
    analytic: boolean
        compute field using analytic formula (True) or quadrature (False)

    Returns
    -------
    C: ndarray (Np, 3, Nvertices)
        Coupling matrix corresponding to a mapping from a stream function
        on the mesh to B-field at the evaluation points

    """

    if analytic:
        return magnetic_field_coupling_analytic(mesh, r, Nchunks)

    mu0 = 4 * np.pi * 1e-7
    coef = mu0 / (4 * np.pi)

    print(
        "Computing magnetic field coupling matrix, %d vertices by %d target points... "
        % (len(mesh.vertices), len(r)),
        end="",
    )
    start = time.time()

    w_quad, r_quad = get_quad_points(
        mesh.vertices, mesh.faces, ("dunavant", quad_degree)
    )

    # Rotated gradients (currents)
    Gx, Gy, Gz = gradient_matrix(mesh, rotated=True)

    # Initialize C-matrix
    n_target_points = len(r)
    n_verts = len(mesh.vertices)
    C = np.zeros((n_target_points, n_verts, 3))

    if Nchunks is None:
        if r.shape[0] > 1000:
            Nchunks = r.shape[0] // 100
        else:
            Nchunks = 1

    for n in range(Nchunks):
        # Diffence vectors (Neval, Ntri, Nquad, 3)
        RR = r_quad[None, :, :, :] - r[n::Nchunks, None, None, :]

        # RR/norm(RR)**3 "Gradient of Green's function"
        g = -RR / ((np.linalg.norm(RR, axis=-1) ** 3)[:, :, :, None])

        # Sum over quad points and multiply by triangle area
        g = (g * w_quad[:, None]).sum(axis=-2)
        g *= mesh.area_faces[:, None]

        # Cross product RR/norm(RR)
        C[n::Nchunks, :, 0] = g[:, :, 2] @ Gy - g[:, :, 1] @ Gz
        C[n::Nchunks, :, 1] = g[:, :, 0] @ Gz - g[:, :, 2] @ Gx
        C[n::Nchunks, :, 2] = g[:, :, 1] @ Gx - g[:, :, 0] @ Gy

    duration = time.time() - start
    print("took %.2f seconds." % duration)

    C *= coef
    #    return np.moveaxis(C, 2, 1)
    return np.swapaxes(C, 2, 1)


def magnetic_field_coupling_analytic(mesh, r, Nchunks=None):
    """
    Computes the coupling matrix which gives the magnetic field at
    target points due to currents (stream function) on a surface mesh using analytical formulas.

    Parameters
    ----------

    mesh: Trimesh mesh object
        mesh describing the geometry of the field source
    r: ndarray (Np, 3)
        evaluation points
    Nchunks: int
        number of chunks used in the calculation for saving memory

    Returns
    -------
    C: ndarray (Np, 3, Nvertices)
        Coupling matrix corresponding to a mapping from a stream function
        on the mesh to B-field at the evaluation points

    """
    from .integrals import omega, gamma0

    coef = 1e-7

    print(
        "Computing magnetic field coupling matrix analytically, %d vertices by %d target points... "
        % (len(mesh.vertices), len(r)),
        end="",
    )
    start = time.time()

    if Nchunks is None:
        if r.shape[0] > 1000:
            Nchunks = r.shape[0] // 100
        else:
            Nchunks = 1

    tn = mesh.face_normals

    # Nfaces, 3, 3
    rfaces = mesh.vertices[mesh.faces]

    # Calculate potentials and related coefficients
    gamma_terms = np.zeros((r.shape[0], mesh.faces.shape[0], 3))
    omega_terms = np.zeros((r.shape[0], mesh.faces.shape[0]))
    # Edges Nfaces, 3, 3
    edges = np.roll(rfaces, 1, -2) - np.roll(rfaces, 2, -2)
    for n in range(Nchunks):
        RRchunk = r[n::Nchunks, None, None, :] - rfaces[None, :, :, :]
        # Neval, Nfaces, xyz
        gamma_terms[n::Nchunks] = -np.einsum("nfe,fei->nfi", gamma0(RRchunk), edges)
        omega_terms[n::Nchunks] = omega(RRchunk)

    # 3 (Nfaces, Nverts) sparse matrices
    G = gradient_matrix(mesh, rotated=False)
    R = gradient_matrix(mesh, rotated=True)
    C = np.zeros((3, r.shape[0], mesh.vertices.shape[0]))

    # Accumulate elements by sparse matrix products for x,y, and z components
    for fcomp in range(3):
        C[fcomp] = omega_terms @ G[fcomp]
        # Accumulate gamma terms for each vertex in the triangle
        for gcomp in range(3):
            # Edge @ Rotated_gradient "==" c_coeff
            # Multiplying with R-matrices takes care of c_coeff calculation
            # and accumulation to right vertex
            C[fcomp] += (tn[:, fcomp] * gamma_terms[:, :, gcomp]) @ R[gcomp]

    duration = time.time() - start
    print("took %.2f seconds." % duration)

    C *= coef
    #    return np.moveaxis(C, 0, 1)
    return np.swapaxes(C, 0, 1)


def scalar_potential_coupling(
    mesh, r, Nchunks=None, multiply_coeff=False, approx_far=False, margin=3
):
    """
    Coupling matrix from a stream function on a mesh
    to scalar potential using analytic integrals.

    Parameters
    ----------

    mesh: Trimesh mesh object
        mesh describing the geometry of the field source
    r: ndarray (Np, 3)
        evaluation points
    Nchunks: int
        number of chunks used in the calculation for saving memory
    multiply_coeff: boolean
        If True, multiply result by mu_0/(4*pi)
    approx_far: boolean,
        approximate the potential using simple quadrature
        (see integrals.potential_dipoles) for points far from the source triangles
    margin: float
        cut-off distance for "far" points measured in mean triangle side length.

    Returns
    -------
    U: ndarray (Np, Nvertices)
        Coupling matrix corresponding to a mapping from a stream function
        on the mesh to scalar potential at the evaluation points
    """

    print(
        "Computing scalar potential coupling matrix, %d vertices by %d target points... "
        % (len(mesh.vertices), len(r)),
        end="",
    )
    start = time.time()

    # Source and eval locations
    R1 = mesh.vertices[mesh.faces]
    R2 = r

    R2chunks, ichunks = get_chunks(R2, Nchunks, True)
    Uf = np.zeros((R2.shape[0], mesh.faces.shape[0], 3))
    far_chunks = []

    for ichunk, R2chunk in zip(ichunks, R2chunks):
        RRchunk = R2chunk[:, None, None, :] - R1[None, :, :, :]
        RRchunk_verts = R2chunk[:, None, :] - mesh.vertices[None, :, :]
        if approx_far:
            temp = np.zeros(RRchunk.shape[:3])
            # near, far = _split_by_distance(mesh, RRchunk, margin)
            near_v, far_v = _split_by_distance(mesh, RRchunk_verts, margin)
            near = mesh.faces_sparse.T @ near_v
            far_chunks.append(far_v)
            # far = np.invert(near)
            temp[:, near, :] = triangle_potential_dipole_linear(
                RRchunk[:, near], mesh.face_normals[near], mesh.area_faces[near]
            )
            # This far approximation does not speed up the computation much
            # because the quadrature points are so many
            # temp[:,far,:] = potential_dipoles(RRchunk[:, far],
            # mesh.face_normals[far],
            # mesh.area_faces[far])
            Uf[ichunk] = temp
        else:
            Uf[ichunk] = triangle_potential_dipole_linear(
                RRchunk, mesh.face_normals, mesh.area_faces
            )

    # Sparse products are equivalent to this
    # Uv = np.zeros((R2.shape[0], mesh.vertices.shape[0]))
    # for ind_f, f in enumerate(mesh.faces):
    #     Uv[:, f] += Uf[:, ind_f]
    from scipy.sparse import csc_matrix

    Nf = len(mesh.faces)
    Nv = len(mesh.vertices)
    M0 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 0])), (Nf, Nv))
    M1 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 1])), (Nf, Nv))
    M2 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 2])), (Nf, Nv))
    Uv = Uf[:, :, 0] @ M0 + Uf[:, :, 1] @ M1 + Uf[:, :, 2] @ M2

    # Calcuate far points by vertex based approximation
    if approx_far:
        areas = mass_matrix(mesh, lumped=True).diagonal()
        for ichunk, R2chunk, far in zip(ichunks, R2chunks, far_chunks):
            RRchunk_verts = R2chunk[:, None, :] - mesh.vertices[None, far, :]
            mask = ichunk[:, None] * far
            Uv[mask] = potential_vertex_dipoles(
                RRchunk_verts, mesh.vertex_normals[far], areas[far]
            ).ravel()

    duration = time.time() - start
    print("took %.2f seconds." % duration)

    if multiply_coeff:
        coeff = 1e-7  # mu_0/(4*pi)
    else:
        coeff = 1 / (4 * np.pi)
    return Uv * coeff


def _triangle_coupling(
    mesh, r, Nchunks=None, approx_far=True, margin=2, chunk_clusters=False, planar=False
):
    """

    Parameters
    ----------
    mesh: Trimesh mesh object
        mesh describing the geometry of the field source
    r: ndarray (Np, 3)
        evaluation points
    Nchunks: int
        number of chunks used in the calculation for saving memory
    approx_far : boolean
        speed up the calculation by approxmating far points. The default is True.
    margin : boolean, optional
        defintion of far points in average triangle sidelength. The default is 2.
    chunk_clusters : boolean, optional
        make chunks clusters, may speed up the calculation. The default is False.
    planar : boolean, optional
        Prapagated to triangle_potential_uniform. For planar meshes
        the calculation can be speeded up. The default is False.

    Returns
    -------
    M : TYPE
        DESCRIPTION.

    """

    # Source and eval locations
    R1 = mesh.vertices[mesh.faces]
    R2 = r

    R2chunks, ichunks = get_chunks(R2, Nchunks, chunk_clusters)

    M = np.zeros((R2.shape[0], mesh.faces.shape[0]))
    print("Computing triangle-coupling matrix")

    if planar:
        print("Assuming the mesh is planar (if not, set planar=False)")

    for ichunk, R2chunk in zip(ichunks, R2chunks):
        RRchunk = R2chunk[:, None, None, :] - R1[None, :, :, :]
        if approx_far:
            RRchunk_centers = R2chunk[:, None, :] - mesh.triangles_center[None, :, :]
            temp = np.zeros(RRchunk.shape[:2])
            near, far = _split_by_distance(mesh, RRchunk_centers, margin)
            temp[:, near] = triangle_potential_uniform(
                RRchunk[:, near], mesh.face_normals[near], planar
            )
            temp[:, far] = triangle_potential_approx(
                RRchunk_centers[:, far], mesh.area_faces[far], reg=0
            )
            M[ichunk] = temp
        else:
            M[ichunk] = triangle_potential_uniform(RRchunk, mesh.face_normals, planar)

    return M


def vector_potential_coupling(
    mesh, r, Nchunks=None, approx_far=True, margin=2, chunk_clusters=False
):
    """
    Compute vector potential coupling matrices
    from a linear stream function on a mesh using analytic integrals.

    Parameters
    ----------

    mesh: Trimesh mesh object
        mesh describing the geometry of the field source
    r: ndarray (Np, 3)
        evaluation points
    approx_far: Boolean (True)
        If True, use approximate calculation for triangles that
        far from the source triangles using a simple quadrature
        (see integrals.triangle_potential_approx)
    margin: float
        Cut-off distance for "far" points measured in mean triangle side length.

    Returns
    -------
    A: ndarray (Np, 3, Nvertices)
        Coupling matrix corresponding to a mapping from a stream function
        on the mesh to vector potential at the evaluation points
    """

    coeff = 1e-7  # mu_0/(4*pi)

    Af = _triangle_coupling(mesh, r, Nchunks, approx_far, margin, chunk_clusters)

    # Rotated gradients (currents)
    Gx, Gy, Gz = gradient_matrix(mesh, rotated=True)
    # Accumulate the elements
    Av = np.array([Af @ Gx, Af @ Gy, Af @ Gz])

    return np.swapaxes(Av * coeff, 0, 1)


def get_chunks(r, Nchunks, clusters=True):
    """Chunk points in 'r' to Nchunks

    r : ndarray (Npoints, 3)
    """
    if Nchunks is None:
        if r.shape[0] > 1000:
            Nchunks = r.shape[0] // 100
        else:
            Nchunks = 1
    if clusters:
        # Voronoi cells of random vertices
        i_samples = np.random.randint(0, r.shape[0], Nchunks)
        dists = np.linalg.norm(r[:, None, :] - r[None, i_samples, :], axis=-1)
        labels = np.argmin(dists, axis=1)
        # indices as boolean arrays
        # Number of unique labels can be smaller than Nchunks if
        # there are vertices without any points in their cells
        ichunks = [labels == label for label in np.unique(labels)]
        rchunks = [r[mask] for mask in ichunks]
    else:
        # Chunk r by array split and get the corresponding indices as slices
        rchunks = np.array_split(r, Nchunks, axis=0)
        lengths = [len(ri) for ri in rchunks]
        inds = np.cumsum([0] + lengths)
        ichunks = [slice(inds[i], inds[i + 1]) for i in range(len(lengths))]

    return rchunks, ichunks


def _split_by_distance(mesh, RR, margin=3):
    avg_sidelength = np.sqrt(4 / np.sqrt(3) * np.mean(mesh.area_faces[::100]))

    RRnorm = np.linalg.norm(RR, axis=-1)
    # near = np.nonzero(np.min(RRnorm, axis=0) < avg_sidelength * margin)[0]
    # far = np.setdiff1d(np.arange(0, len(mesh.faces)), near, assume_unique=True)
    near = np.min(RRnorm, axis=0) < avg_sidelength * margin
    far = np.invert(near)

    return near, far


# def _split_by_distance(mesh, RR, margin=3):
#     avg_sidelength = np.sqrt(4/np.sqrt(3)*np.mean(mesh.area_faces[::100]))
# #    np.mean(np.linalg.norm(np.diff(mesh.vertices[mesh.edges[::1000]], axis=1), axis=-1))

#     RRnorm = np.linalg.norm(RR, axis=-1)
#     near = np.nonzero(np.min(RRnorm, axis=(0, 2)) < avg_sidelength * margin)[0]

#     far = np.setdiff1d(np.arange(0, len(mesh.faces)), near, assume_unique=True)

# #    print('near: %d, far: %d'%(len(near), len(far)))
#     return near, far



import numpy as np
import matplotlib.pyplot as plt
import trimesh
from mayavi import mlab
from bfieldtools.thermal_noise import (
    compute_current_modes,
    noise_covar,
    noise_var,
    visualize_current_modes,
)
from bfieldtools.mesh_magnetics import magnetic_field_coupling

import pkg_resources


font = {"family": "normal", "weight": "normal", "size": 20}
plt.rc("font", **font)

# Fix the simulation parameters
d = 1e-3
sigma = 3.8e7
T = 293
kB = 1.38064852e-23
mu0 = 4 * np.pi * 1e-7
freqs = np.array((0,))


Nchunks = 8
quad_degree = 2

Nfreqs = 70
freqs = np.linspace(0, 200, Nfreqs)

#%%
mesh = trimesh.load(
    pkg_resources.resource_filename(
        "bfieldtools", "example_meshes/unitdisc_extremelyfine.stl"
    )
)

vl = compute_current_modes(
    obj=mesh,
    T=T,
    resistivity=1 / sigma,
    thickness=d,
    mode="AC",
    freqs=freqs,
    return_eigenvals=False,
)

#%%
Np = 25
z = np.linspace(0.05, 1, Np)
fp = np.array((np.zeros(z.shape), np.zeros(z.shape), z)).T

B_coupling = magnetic_field_coupling(mesh, fp, analytic=True)

Bf = np.sqrt(noise_var(B_coupling, vl))

plt.figure(figsize=(5, 5))
plt.loglog(freqs, Bf[:, 2, :].T * 1e15, linewidth=2)
plt.grid()
plt.ylim(0.1, 100)
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False)
plt.xlabel("Frequency (Hz)")
plt.ylabel(r"$B_z$ noise (fT/rHz)")
plt.tight_layout()

f_interp = np.linspace(0, 120, 500)

cutf = np.zeros(Np)
for i in range(Np):
    Btemp = np.interp(f_interp, freqs, Bf[i, 2, :])
    idx = np.max(np.where(Btemp >= 1 / np.sqrt(2) * Btemp[0]))
    cutf[i] = f_interp[idx]

cutf_an = 1 / (4 * mu0 * sigma * d * z)

plt.figure(figsize=(5, 5))
plt.loglog(z, cutf_an, linewidth=2, label="Infinite plane")
plt.loglog(z, cutf, "x", markersize=10, markeredgewidth=2, label="Disc")
plt.grid()
plt.gca().spines["right"].set_visible(False)
plt.gca().spines["top"].set_visible(False)
plt.legend(frameon=False, loc=3)
plt.xlabel("Distance (z/R)")
plt.ylabel("3-dB cutoff frequency (Hz)")
plt.tight_layout()



# -*- coding: utf-8 -*-
"""
Created on Wed Apr  8 15:09:56 2020

@author: Rasmus Zetter
"""

import pytest
from bfieldtools import viz
from bfieldtools.utils import load_example_mesh

from .test_mesh_conductor import _fake_mesh_conductor, _fake_streamfunction


import numpy as np
from mayavi import mlab
import matplotlib.pyplot as plt


def test_mesh_plotting():

    mesh = load_example_mesh("unit_disc")

    v_scalars = np.random.rand(len(mesh.vertices))
    f_scalars = np.random.rand(len(mesh.faces))

    viz.plot_mesh(mesh, figure=None)
    f = mlab.figure()
    viz.plot_mesh(mesh, figure=f)

    viz.plot_data_on_faces(mesh, f_scalars)

    f = mlab.figure()
    viz.plot_data_on_faces(mesh, f_scalars, figure=f, colormap="jet")

    f_scalars = np.random.rand(len(mesh.faces)) - 0.5
    viz.plot_data_on_faces(mesh, f_scalars, figure=f, colorbar=True)

    viz.plot_data_on_vertices(mesh, v_scalars, colorbar=True)
    viz.plot_data_on_vertices(mesh, v_scalars, autoscale=True)

    v_scalars = np.random.rand(len(mesh.vertices)) - 0.5

    viz.plot_data_on_vertices(mesh, v_scalars, colormap="jet")


def test_plot_cross_section():

    X, Y = np.meshgrid(np.linspace(0, 1, 11), np.linspace(0, 1, 11))
    Z = 3 - (X**2 + Y**2)

    viz.plot_cross_section(X, Y, Z**10, log=True, colorbar=True, cmap=None)
    viz.plot_cross_section(X, Y, Z, log=False, cmap=None)
    viz.plot_cross_section(
        X, Y, Z, log=False, cmap="jet", vmin=-10, vmax=10, contours=False
    )

    X, Y = np.meshgrid(np.linspace(0, 1, 11), np.linspace(0, 1, 11))
    Z = 3 - 9 * (X**2 + Y**2)

    viz.plot_cross_section(X, Y, Z, log=False, cmap=None, contours=4)

    fig, axes = plt.subplots(1, 1)
    viz.plot_cross_section(X, Y, Z, log=False, axes=axes, vmin=None, vmax=None)


def test_plot_3d_current_loops():

    x = np.cos(np.linspace(0, 2 * np.pi))
    y = np.sin(np.linspace(0, 2 * np.pi))
    z = np.zeros_like(y)
    loops = [np.array([x, y, z]).T, 2 * np.array([x, y, z]).T]

    viz.plot_3d_current_loops(loops, tube_radius=0.01)
    viz.plot_3d_current_loops(loops, tube_radius=0.01, colors="auto")
    f = mlab.figure()
    viz.plot_3d_current_loops(loops, tube_radius=0.01, figure=f)

    viz.plot_3d_current_loops(loops, tube_radius=0.01, colors=None)
    viz.plot_3d_current_loops(loops, tube_radius=0.01, colors=(1, 0, 0))
    viz.plot_3d_current_loops(loops, tube_radius=0.01, colors=[(1, 0, 0)] * 2)

    # Longest segment is the one closing the loop
    x = np.cos(np.linspace(0, 2 * np.pi))[:-1]
    y = np.sin(np.linspace(0, 2 * np.pi))[:-1]
    z = np.zeros_like(y)
    loops = [np.array([x, y, z]).T, 2 * np.array([x, y, z]).T]
    viz.plot_3d_current_loops(loops, tube_radius=0.01)


@pytest.mark.xfail
def test_plot_3d_current_loops_fail():

    x = np.cos(np.linspace(0, 2 * np.pi))
    y = np.sin(np.linspace(0, 2 * np.pi))
    z = np.zeros_like(y)
    loops = [np.array([x, y, z]).T, 2 * np.array([x, y, z]).T]

    viz.plot_3d_current_loops(loops, tube_radius=0.01, colors="boop")



"""
Visualization functions tailored for bfieldtools.
Mainly wrappers and convenience helpers around mayavi and matplotlib functions
"""

__all__ = [
    "plot_3d_current_loops",
    "plot_cross_section",
    "plot_data_on_faces",
    "plot_data_on_vertices",
    "plot_mesh",
]

import matplotlib.pyplot as plt
from matplotlib import colors as c

import numpy as np


def plot_mesh(
    mesh, cull_front=False, cull_back=False, figure=None, figsize=(800, 800), **kwargs
):
    """
    Plot the mesh surface in mayavi.

    Parameters
    ----------
    mesh: Trimesh mesh object
        mesh to be plotted
    cull_front: Boolean (False)
        If True, cull front of mesh
    cull_back: Boolean (False)
        If True, cull back of mesh
    figure: mayavi figure or None
        If passed, plot to existing figure
    figsize: tuple (x, y)
        Figure size (if figure is None)

    Returns
    -------
    figure: mayavi figure
        Contains the plotted mesh

    """
    from mayavi import mlab

    if figure is None:
        figure = mlab.figure(
            None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=figsize
        )

    meshviz = mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, **kwargs)

    meshviz.actor.mapper.scalar_visibility = False

    meshviz.actor.property.frontface_culling = cull_front
    meshviz.actor.property.backface_culling = cull_back
    return figure


def plot_3d_current_loops(
    current_loops,
    colors="auto",
    figure=None,
    figsize=(800, 800),
    tube_radius=0.05,
    origin=np.array([0, 0, 0]),
):
    """
    Plot current loops (e.g. contour_polys given by scalar_contour()) in 3D using mayavi.

    Parameters
    ----------
    current_loops: list of (N_verts, 3) arrays with length N_loops
        List of segmented lines specifying the path of a current loop.
        Last vertex is assumed to be connected to the first.

    colors: 'auto' (default) or None or (3, ) tuple or  list of (3, ) tuples with same length as current_loops
        Optional specification of colors used for current direction. If auto (default),
        color blue-red based on current direction. If None, use middle-grey for all loops.
        If tuple, color all loops according to tuple. If list of tuples, color each loop accordingly.
    origin: array-like with length 3
        Shifts the origin used to determine the 'auto' coloring.
    tune_radius: float
        radius of loops plotted
    figure: existing mlab figure
        Optional, if passed will plot to existing figure
    figsize: (x, y) tuple
        Optional, if plotting to new figure specifies the size (in pixels)

    Returns
    -------
    fig: mlab figure

    """
    from mayavi import mlab

    if figure is None:
        figure = mlab.figure(
            None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=figsize
        )

    if colors is None:
        colors = [(0.5, 0.5, 0.5)] * len(current_loops)

    elif isinstance(colors, tuple):
        colors = [colors] * len(current_loops)

    elif isinstance(colors, list):
        assert len(colors) == len(current_loops)

    elif colors == "auto":
        colors = []

        palette = [(1, 0, 0), (0, 0, 1)]
        for loop_idx, loop in enumerate(current_loops):

            # Compute each loop segment
            segments = np.vstack(
                (loop[1:, :] - loop[0:-1, :], loop[0, :] - loop[-1, :])
            )

            # Find mean normal vector following right-hand rule, in loop centre
            centre_normal = np.mean(np.cross(segments, loop), axis=0)
            centre_normal /= np.linalg.norm(centre_normal, axis=-1)

            # Check if normal "points in" or "out" (towards or away from origin)
            origin_vector = np.mean(loop, axis=0) - origin

            colors.append(
                palette[int((np.sign(centre_normal @ origin_vector) + 1) / 2)]
            )
    else:
        raise ValueError("Invalid parameter for colors")

    for loop_idx, loop in enumerate(current_loops):
        mlab.plot3d(
            *loop[list(range(len(loop))) + [0]].T,
            color=colors[loop_idx],
            tube_radius=tube_radius
        )

        # Put two arrows on loop

        # First arrow on longest segment of loop
        longest_idx = np.argmax(
            np.linalg.norm(
                np.vstack((loop[1:, :] - loop[0:-1, :], loop[0, :] - loop[-1, :])),
                axis=-1,
            )
        )

        if longest_idx == len(loop) - 1:
            arrow1 = mlab.quiver3d(
                *loop[-1, :].T,
                *(loop[0, :] - loop[-1, :]).T,
                mode="cone",
                scale_mode="none",
                scale_factor=0.5 * tube_radius / 0.05,
                color=colors[loop_idx]
            )
        else:
            arrow1 = mlab.quiver3d(
                *loop[longest_idx + 1, :].T,
                *(loop[longest_idx + 1, :] - loop[longest_idx, :]).T,
                mode="cone",
                scale_mode="none",
                scale_factor=0.5 * tube_radius / 0.05,
                color=colors[loop_idx]
            )

        arrow1.glyph.glyph_source.glyph_position = "center"
        arrow1.glyph.glyph_source.glyph_source.radius = 0.3
        arrow1.glyph.glyph_source.glyph_source.height = 0.5

    #        #Second arrow on the element "half away"
    #
    #        opposite_idx = int((longest_idx + len(loop)/2) % len(loop))
    #
    #        if opposite_idx == len(loop)-1:
    #            arrow1 = mlab.quiver3d(*loop[-1,:].T,
    #                      *(loop[0,:] - loop[-1,:]).T,
    #                      mode='cone', scale_mode='none',
    #                      scale_factor=0.5 * tube_radius/0.05,
    #                      color=colors[loop_idx])
    #        else:
    #            arrow2 = mlab.quiver3d(*loop[opposite_idx+1,:].T,
    #                      *(loop[opposite_idx+1,:] - loop[opposite_idx,:]).T,
    #                      mode='cone', scale_mode='none',
    #                      scale_factor=0.5 * tube_radius/0.05,
    #                      color=colors[loop_idx])
    #        arrow2.glyph.glyph_source.glyph_position = 'center'
    #        arrow2.glyph.glyph_source.glyph_source.radius = 0.3
    #        arrow2.glyph.glyph_source.glyph_source.height = 0.5

    figure.scene.isometric_view()

    return figure


def plot_cross_section(
    X,
    Y,
    data,
    axes=None,
    cmap=None,
    colorbar=False,
    contours=10,
    log=False,
    vmin=None,
    vmax=None,
):
    """
    Plot scalar data on a plane

    Parameters
    ----------
    X: (N_x, N_y)
        meshgrid-generated 2D array with X coordinates for each data point
    Y: (N_x, N_y) array
        meshgrid-generated 2D array with Y coordinates for each data point
    data: (N_x, N_y) array
        2D array with data for each point in X, Y
    axes: matplotlib axes
        Optional, if passed plot to existing axes. If None (default), plot to new figure/axes
    cmap: str
        name of colormap to use for scalar data. If None (default), use viridis for
        all-positive/-negative data and RdBu otherwise
    colorbar: Boolean
        If True, plot colorbar for data
    contours: None, int or array-like
        If None, no contour lines are plotted. If int, plots specific number of
        equispaced contour lines. If array-like, plot contours at values in array.
    log: Boolean
        if True, colormap is log-scaled. Else, colormap is linear.
    vmin, vmax: float or None
        Explicit colormap minimum and maximum values. If not passed, range is according to data

    Returns
    -------
    axes: matplotlib axes with plot
    """

    if axes is None:
        fig, axes = plt.subplots(1, 1)

    # If data is all-positive or all-negative, use viridis. Otherwise, use Red-Blue colormap
    if cmap is None:
        if np.all(data > 0) or np.all(data < 0):
            cmap = "viridis"
        else:
            cmap = "RdBu"

    if vmin is None:
        vmin = np.min(data)

    if vmax is None:
        vmax = np.max(data)

    if log:
        norm = c.LogNorm(vmin, vmax)
    else:
        norm = c.Normalize(vmin, vmax)

    cont = axes.pcolormesh(X, Y, data, cmap=cmap, norm=norm, shading="gouraud")

    if colorbar:
        plt.colorbar(cont)

    if contours:
        clines = axes.contour(
            X,
            Y,
            data,
            levels=contours,
            norm=norm,
            antialiased=True,
            colors=("k",),
            linewidths=(1,),
        )

        axes.clabel(clines, fmt="%2.2f", colors="w", fontsize=10)

    axes.set_xlabel("X")
    axes.set_ylabel("Y")

    axes.figure.tight_layout()

    return axes


def plot_data_on_faces(
    mesh,
    data,
    figure=None,
    figsize=(800, 800),
    colorbar=False,
    ncolors=256,
    vmin=None,
    vmax=None,
    **kwargs
):
    """Plot any data determined on the faces of a mesh

    Parameters
    ----------
    mesh: Trimesh mesh object
    data: (N_verts, ) array
        Scalar data to plot on mesh faces
    figure: existing mlab figure
        Optional, if passed will plot to existing figure
    figsize: (x, y) tuple
        Optional, if plotting to new figure specifies the size (in pixels)
    colormap: str
        name of colormap to use for scalar data. If None (default), use viridis for
        all-positive/-negative data and RdBu otherwise
    colorbar: Boolean
        If True, plot colorbar for scalar data
    ncolors: int
        Number of colors to use

    Returns
    -------
    fig: mlab figure
    """
    from mayavi import mlab

    if figure is None:
        figure = mlab.figure(
            None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=figsize
        )

    # If data is all-positive or all-negative, use viridis. Otherwise, use Red-Blue colormap
    if "colormap" not in kwargs:
        if np.all(data > 0) or np.all(data < 0):
            kwargs["colormap"] = "viridis"
        else:
            kwargs["colormap"] = "RdBu"

    v = mesh.vertices
    f = mesh.faces

    s = mlab.pipeline.triangular_mesh_source(*v.T, f)
    s.mlab_source.dataset.cell_data.scalars = data

    s.mlab_source.dataset.cell_data.scalars.name = "Cell data"

    s.mlab_source.update()
    s2 = mlab.pipeline.set_active_attribute(s, cell_scalars="Cell data")
    surf = mlab.pipeline.surface(s2, **kwargs)

    if colorbar:
        mlab.colorbar(surf)

    lutmanager = surf.parent.scalar_lut_manager
    lutmanager.lut_mode = kwargs["colormap"]

    if kwargs["colormap"] == "RdBu":
        rangemax = np.max(np.abs(data))
        if vmin is None:
            vmin = -rangemax * 1.01
        if vmax is None:
            vmax = rangemax * 1.01
        lutmanager.data_range = np.array([vmin, vmax])
        lutmanager.use_default_range = False

    lutmanager.number_of_colors = ncolors

    return surf


def plot_data_on_vertices(
    mesh,
    data,
    figure=None,
    figsize=(800, 800),
    colorbar=False,
    ncolors=256,
    interpolate=True,
    cull_front=False,
    cull_back=False,
    autoscale=False,
    vmin=None,
    vmax=None,
    **kwargs
):
    """
    Plot scalar data defined on the vertices of a mesh.

    Parameters
    ----------
    mesh: Trimesh mesh object
    data: (N_verts, ) array
        Scalar data to plot on mesh vertices
    figure: existing mlab figure
        Optional, if passed will plot to existing figure
    figsize: (x, y) tuple
        Optional, if plotting to new figure specifies the size (in pixels)
    cmap: str
        name of colormap to use for scalar data. If None (default), use viridis for
        all-positive/-negative data and RdBu otherwise
    colorbar: Boolean
        If True, plot colorbar for scalar data
    ncolors: int
        Number of colors to use
    interpolate: Boolean
        If True, interpolate scalar data for smoother look
    opacity: float
        Opacity of rendered mesh

    Returns
    -------
    fig: mlab figure

    """
    from mayavi import mlab

    if figure is None:
        figure = mlab.figure(
            None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=figsize
        )

    # If data is all-positive or all-negative, use viridis. Otherwise, use Red-Blue colormap
    if "colormap" not in kwargs:
        if np.all(data > 0) or np.all(data < 0):
            kwargs["colormap"] = "viridis"
        else:
            kwargs["colormap"] = "RdBu"

    surf = mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, scalars=data, **kwargs)
    surf.actor.property.frontface_culling = cull_front
    surf.actor.property.backface_culling = cull_back
    if colorbar:
        mlab.colorbar(surf)
    surf.actor.mapper.interpolate_scalars_before_mapping = interpolate

    lutmanager = surf.module_manager.scalar_lut_manager
    lutmanager.number_of_colors = ncolors

    if (np.all(data > 0) or np.all(data < 0)) and autoscale:
        rangemax = np.max(np.abs(data))
        lutmanager.data_range = np.array([-rangemax * 1.01, rangemax * 1.01])
        lutmanager.use_default_range = False
    elif vmax is not None:
        if vmin is None:
            lutmanager.data_range = np.array([-vmax, vmax])
        else:
            lutmanager.data_range = np.array([vmin, vmax])
        lutmanager.use_default_range = False

    return surf



"""
Magnetically shielded  coil
===========================
Compact example of design of a biplanar coil within a cylindrical shield.
The effect of the shield is prospectively taken into account while designing the coil.
The coil is positioned close to the end of the shield to demonstrate the effect
"""

PLOT = True
SAVE_FIGURES = False
SAVE_STREAMFUNCTION = False

SAVE_DIR = "./Shielded coil/"

import numpy as np
from mayavi import mlab
import trimesh


from bfieldtools.mesh_conductor import MeshConductor, StreamFunction
from bfieldtools.coil_optimize import optimize_streamfunctions
from bfieldtools.utils import load_example_mesh, combine_meshes

from bfieldtools.contour import scalar_contour
from bfieldtools.viz import plot_3d_current_loops

import pkg_resources


# Set unit, e.g. meter or millimeter.
# This doesn't matter, the problem is scale-invariant
scaling_factor = 1


# Load simple plane mesh that is centered on the origin
planemesh = load_example_mesh("10x10_plane_hires")
planemesh.apply_scale(scaling_factor)

# Specify coil plane geometry
center_offset = np.array([9, 0, 0]) * scaling_factor
standoff = np.array([0, 4, 0]) * scaling_factor

# Create coil plane pairs
coil_plus = trimesh.Trimesh(
    planemesh.vertices + center_offset + standoff, planemesh.faces, process=False
)

coil_minus = trimesh.Trimesh(
    planemesh.vertices + center_offset - standoff, planemesh.faces, process=False
)

joined_planes = combine_meshes((coil_minus, coil_plus))


# Create mesh class object
coil = MeshConductor(mesh_obj=joined_planes, fix_normals=True, basis_name="inner")

# Separate object for shield geometry
shieldmesh = trimesh.load(
    file_obj=pkg_resources.resource_filename(
        "bfieldtools", "example_meshes/closed_cylinder_remeshed.stl"
    ),
    process=True,
)
shieldmesh.apply_scale(15)

shield = MeshConductor(
    mesh_obj=shieldmesh, process=True, fix_normals=True, basis_name="vertex"
)


#%%
# Set up target  points and plot geometry

# Here, the target points are on a volumetric grid within a sphere
# Set up target and stray field points

# Here, the target points are on a volumetric grid within a sphere

center = np.array([9, 0, 0]) * scaling_factor

sidelength = 3 * scaling_factor
n = 12
xx = np.linspace(-sidelength / 2, sidelength / 2, n)
yy = np.linspace(-sidelength / 2, sidelength / 2, n)
zz = np.linspace(-sidelength / 2, sidelength / 2, n)
X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

x = X.ravel()
y = Y.ravel()
z = Z.ravel()

target_points = np.array([x, y, z]).T

# Turn cube into sphere by rejecting points "in the corners"
target_points = (
    target_points[np.linalg.norm(target_points, axis=1) < sidelength / 2] + center
)


# Plot coil, shield and target points
if PLOT:
    f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))

    coil.plot_mesh(representation="surface", figure=f, opacity=0.5)
    shield.plot_mesh(representation="surface", opacity=0.2, figure=f)
    mlab.points3d(*target_points.T)

    f.scene.isometric_view()
    f.scene.camera.zoom(1.1)

    if SAVE_FIGURES:
        mlab.savefig(
            SAVE_DIR + "shielded_biplanar_geometry.png",
            figure=f,
            magnification=4,
        )
        mlab.close()


#%%
# Let's design a coil without taking the magnetic shield into account

# The absolute target field amplitude is not of importance,
# and it is scaled to match the C matrix in the optimization function
target_field = np.zeros(target_points.shape)
target_field[:, 0] = target_field[:, 0] + 1  # Homogeneous Y-field


target_abs_error = np.zeros_like(target_field)
target_abs_error[:, 0] += 0.005
target_abs_error[:, 1:3] += 0.01

target_spec = {
    "coupling": coil.B_coupling(target_points),
    "abs_error": target_abs_error,
    "target": target_field,
}

import mosek

coil.s, coil.prob = optimize_streamfunctions(
    coil,
    [target_spec],
    objective="minimum_inductive_energy",
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)


if SAVE_STREAMFUNCTION:
    np.save(SAVE_DIR + "biplanar_streamfunction.npy", coil.s.vert)
#%%
# Plot coil windings and target points

loops = scalar_contour(coil.mesh, coil.s.vert, N_contours=10)

if PLOT:
    f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
    mlab.clf()

    plot_3d_current_loops(loops, colors="auto", figure=f)

    B_target = coil.B_coupling(target_points) @ coil.s

    mlab.quiver3d(*target_points.T, *B_target.T, mode="arrow", scale_factor=0.75)

    f.scene.isometric_view()
    f.scene.camera.zoom(0.95)

    if SAVE_FIGURES:
        mlab.savefig(
            SAVE_DIR + "shielded_biplanar_ignored.png",
            figure=f,
            magnification=4,
        )
        mlab.close()
#%%
# Now, let's compute the effect of the shield on the field produced by the coil

# Points slightly inside the shield
d = (
    np.mean(np.diff(shield.mesh.vertices[shield.mesh.faces[:, 0:2]], axis=1), axis=0)
    / 10
)
points = shield.mesh.vertices - d * shield.mesh.vertex_normals


# Solve equivalent stream function for the perfect linear mu-metal layer.
# This is the equivalent surface current in the shield that would cause its
# scalar magnetic potential to be constant
shield.s = StreamFunction(
    np.linalg.solve(shield.U_coupling(points), coil.U_coupling(points) @ coil.s), shield
)

#%%
# Plot the difference in field when taking the shield into account

if PLOT:
    f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
    mlab.clf()

    B_target = coil.B_coupling(target_points) @ coil.s

    B_target_w_shield = (
        coil.B_coupling(target_points) @ coil.s
        + shield.B_coupling(target_points) @ shield.s
    )

    B_quiver = mlab.quiver3d(
        *target_points.T,
        *(B_target_w_shield - B_target).T,
        colormap="viridis",
        mode="arrow"
    )
    f.scene.isometric_view()
    mlab.colorbar(B_quiver, title="Difference in magnetic field (a.u.)")

#%%
# Let's redesign the coil taking the shield into account prospectively

shield.coupling = np.linalg.solve(shield.U_coupling(points), coil.U_coupling(points))

secondary_C = shield.B_coupling(target_points) @ shield.coupling

total_C = coil.B_coupling(target_points) + secondary_C

target_spec_w_shield = {
    "coupling": total_C,
    "abs_error": target_abs_error,
    "target": target_field,
}


coil.s2, coil.prob2 = optimize_streamfunctions(
    coil,
    [target_spec_w_shield],
    objective="minimum_inductive_energy",
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)

#%%
# Plot the newly designed coil windings and field at the target points

loops = scalar_contour(coil.mesh, coil.s2.vert, N_contours=10)

if PLOT:
    f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
    mlab.clf()

    plot_3d_current_loops(loops, colors="auto", figure=f)

    B_target2 = total_C @ coil.s2
    mlab.quiver3d(*target_points.T, *B_target2.T, mode="arrow", scale_factor=0.75)

    f.scene.isometric_view()
    f.scene.camera.zoom(0.95)
    if SAVE_FIGURES:
        mlab.savefig(
            SAVE_DIR + "shielded_biplanar_prospective.png",
            figure=f,
            magnification=4,
        )
        mlab.close()

#%%
# Plot difference in field


import seaborn as sns
import matplotlib.pyplot as plt


if PLOT:
    fig, axes = plt.subplots(1, 3, figsize=(12, 3))

    axnames = ["X", "Y", "Z"]

    # fig.suptitle('Component-wise effect of magnetic shield on target field amplitude distribution')
    for ax_idx, ax in enumerate(axes):

        sns.kdeplot(
            B_target[:, ax_idx],
            label="Coil without shield",
            ax=ax,
            shade=True,
            legend=False,
        )
        sns.kdeplot(
            B_target_w_shield[:, ax_idx],
            label="Coil with shield",
            ax=ax,
            shade=True,
            legend=False,
        )
        sns.kdeplot(
            B_target2[:, ax_idx],
            label="Coil designed with shield",
            ax=ax,
            shade=True,
            legend=False,
        )
        #    ax.set_title(axnames[ax_idx])
        ax.get_yaxis().set_visible(False)

        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["left"].set_visible(False)

        ax.set_xlabel("Magnetic field on %s-axis" % axnames[ax_idx])

        if ax_idx == 0:
            ax.legend()

    fig.tight_layout(rect=[0, 0.03, 1, 0.95])

    if SAVE_FIGURES:
        plt.savefig(SAVE_DIR + "shielding_effect.pdf")



"""
Inductance computation time and memory benchmark
================================================
Benchmark for inductance matrix computation, used
to set auto-chunking.

"""


SAVE = False


import numpy as np
import matplotlib.pyplot as plt
import trimesh

from memory_profiler import memory_usage


def MakeFacesVectorized1(Nr, Nc):

    out = np.empty((Nr - 1, Nc - 1, 2, 3), dtype=int)

    r = np.arange(Nr * Nc).reshape(Nr, Nc)

    out[:, :, 0, 0] = r[:-1, :-1]
    out[:, :, 1, 0] = r[:-1, 1:]
    out[:, :, 0, 1] = r[:-1, 1:]

    out[:, :, 1, 1] = r[1:, 1:]
    out[:, :, :, 2] = r[1:, :-1, None]

    out.shape = (-1, 3)
    return out


Nr = 50
Nc = 50
d = 1


from bfieldtools.mesh_impedance import self_inductance_matrix
import time

N_vertices = []
comp_time = []
mem_use = []


# NE = [20, 30, 40, 60]#, 80]
NE = [20, 25, 30, 40]
for Ne in NE:
    x0 = np.arange(Ne) * d
    y0 = np.arange(Ne) * d
    X, Y = np.meshgrid(x0, y0)
    Z = np.zeros_like(X)

    vertices = np.array([X.flatten(), Y.flatten(), Z.flatten()]).T
    faces = MakeFacesVectorized1(Ne, Ne)

    mesh = trimesh.Trimesh(vertices=vertices, faces=faces)

    N_vertices.append(mesh.vertices.shape[0])
    Nchunks = 1

    #    if mesh.vertices.shape[0] > 3000:
    #        Nchunks=3
    #    if mesh.vertices.shape[0] > 6000:
    #        Nchunks=5
    start_t = time.time()
    mem_use.append(
        np.max(
            memory_usage(
                (
                    self_inductance_matrix,
                    (mesh,),
                    {"Nchunks": Nchunks, "quad_degree": 2},
                )
            )
        )
    )

    comp_time.append(time.time() - start_t)


fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(5, 4))
ax.loglog(N_vertices, comp_time, ".k")
ax.set_xlabel("Number of mesh vertices")
# plt.legend()
ax.set_ylabel("Computation time (s)")
#
# ax.spines['top'].set_visible(False)
# ax.spines['right'].set_visible(False)

ax.grid(which="both", alpha=0.5)

fig.tight_layout()

coefs = np.polyfit(np.log(N_vertices), np.log(comp_time), 1)


N = np.linspace(100, 10000, 200)
T = np.exp(coefs[1]) * N ** coefs[0]

ax.loglog(N, T, "-k", alpha=0.1)
print("Fit:")
print("t = %.6f n**%.2f" % (np.exp(coefs[1]), coefs[0]))

if SAVE:
    fig.savefig("inductance_computation_time.pdf")


fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(5, 4))
ax.loglog(N_vertices, mem_use, ".k")
ax.set_xlabel("Number of mesh vertices")
# plt.legend()
ax.set_ylabel("Memory usage (MiB)")
#
# ax.spines['top'].set_visible(False)
# ax.spines['right'].set_visible(False)

ax.grid(which="both", alpha=0.5)

fig.tight_layout()

mem_coefs = np.polyfit(np.log(N_vertices), np.log(mem_use), 1)


MEM = np.exp(mem_coefs[1]) * N ** mem_coefs[0]

ax.loglog(N, MEM, "-k", alpha=0.1)
print("Fit:")
print("memory = %.6f n**%.2f" % (np.exp(mem_coefs[1]), mem_coefs[0]))

if SAVE:
    fig.savefig("inductance_memory_usage.pdf")



from bfieldtools import utils

import numpy as np

import pytest


def test_mesh_utils():

    mesh1 = utils.load_example_mesh("unit_disc")
    mesh2 = utils.load_example_mesh("10x10_plane")

    mesh2.vertices += np.array([[0, 0, 20]])

    mesh3 = utils.combine_meshes((mesh1, mesh2))

    assert len(mesh3.vertices) == len(mesh1.vertices) + len(mesh2.vertices)


@pytest.mark.xfail
def test_load_mesh_fail():
    mesh = utils.load_example_mesh("this mesh does not exist")


def test_cylinder_points():

    cp = utils.cylinder_points()
    cp = utils.cylinder_points(alpha=90)
    cp = utils.cylinder_points(orientation=np.array([0, 1, 0]))



"""
Figure 1: Stream function example
==================================
    
An example stream function (red--blue colormap) and its rotated gradient, i.e. the surface current density (arrows; green colormap) on a surface mesh with a hole in it. The surface normal is oriented up towards the reader.
"""

from bfieldtools.mesh_conductor import MeshConductor, StreamFunction
import pkg_resources
from bfieldtools.mesh_calculus import gradient
import numpy as np

SAVE = True
SAVE_DIR = "./Streamfunction gradient/"


c = MeshConductor(
    mesh_file=pkg_resources.resource_filename(
        "bfieldtools", "example_meshes/curved_surf_w_hole.stl"
    ),
    process=True,
    basis_name="suh",
    N_suh=10,
    fix_normals=True,
)


T_x = 1.5 * np.pi / 2
T_z = -1.02 * np.pi
rotmat = np.array(
    [
        [np.cos(T_z), -np.sin(T_z), 0, 0],
        [np.sin(T_z), np.cos(T_z), 0, 0],
        [0, 0, 1, 0],
        [0, 0, 0, 1],
    ]
) @ np.array(
    [
        [1, 0, 0, 0],
        [0, np.cos(T_x), -np.sin(T_x), 0],
        [0, np.sin(T_x), np.cos(T_x), 0],
        [0, 0, 0, 1],
    ]
)


c.mesh.apply_transform(rotmat)

s = np.zeros((c.basis.shape[1],))
s[2] += 1
# s[63] += 2

s = StreamFunction(s, c)


from mayavi import mlab
from mayavi.api import Engine

engine = Engine()
engine.start()

f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 700))
s.plot(figure=f, ncolors=256)
c.plot_mesh(representation="wireframe", figure=f)


j = gradient(s.vert, c.mesh, rotated=True)

Len = np.log(np.linalg.norm(j, axis=0))

vectors = mlab.quiver3d(
    *c.mesh.triangles_center.T, *j, mode="arrow", colormap="Greens", scalars=Len
)

# vectors = engine.scenes[0].children[2].children[0].children[0]
vectors.glyph.glyph.scale_mode = "scale_by_scalar"
vectors.glyph.glyph.scale_factor = 0.6
f.scene.z_plus_view()

#
module_manager2 = vectors.module_manager
# module_manager2.scalar_lut_manager.scalar_bar_representation.maximum_size = np.array(
#    [100000, 100000]
# )
# module_manager2.scalar_lut_manager.scalar_bar_representation.minimum_size = np.array(
#    [1, 1]
# )
# module_manager2.scalar_lut_manager.scalar_bar_representation.position = np.array(
#    [0.82, 0.1]
# )
# module_manager2.scalar_lut_manager.scalar_bar_representation.position2 = np.array(
#    [0.17, 0.8]
# )
# module_manager2.scalar_lut_manager.show_scalar_bar = True
# module_manager2.scalar_lut_manager.show_legend = True
# module_manager2.scalar_lut_manager.scalar_bar.height = 0.8
# module_manager2.scalar_lut_manager.scalar_bar.position = np.array([0.82, 0.1])
# module_manager2.scalar_lut_manager.scalar_bar.position2 = np.array([0.17, 0.8])
# module_manager2.scalar_lut_manager.scalar_bar.width = 0.17

f.scene.camera.position = [
    -0.3696892487983681,
    0.2840788710848503,
    3.701830880912346,
]
f.scene.camera.focal_point = [
    -0.3696892487983681,
    0.2840788710848503,
    0.8575533408480627,
]
f.scene.camera.view_angle = 30.0
f.scene.camera.view_up = [0.0, 1.0, 0.0]
f.scene.camera.clipping_range = [2.5164461179149695, 3.263810326333801]
f.scene.camera.compute_view_plane_normal()
f.scene.render()
module_manager2.scalar_lut_manager.scalar_bar.position = np.array([0.82, 0.1])
module_manager2.scalar_lut_manager.scalar_bar.position2 = np.array([0.17, 0.8])
module_manager2.scalar_lut_manager.scalar_bar.number_of_labels = 0
module_manager2.scalar_lut_manager.scalar_bar.position = np.array([0.82, 0.1])
module_manager2.scalar_lut_manager.scalar_bar.position2 = np.array([0.17, 0.8])
module_manager2.scalar_lut_manager.number_of_labels = 0
module_manager2.scalar_lut_manager.scalar_bar.position = np.array([0.82, 0.1])
module_manager2.scalar_lut_manager.scalar_bar.position2 = np.array([0.17, 0.8])
module_manager2.scalar_lut_manager.scalar_bar.maximum_number_of_colors = 8
module_manager2.scalar_lut_manager.scalar_bar.position = np.array([0.82, 0.1])
module_manager2.scalar_lut_manager.scalar_bar.position2 = np.array([0.17, 0.8])
module_manager2.scalar_lut_manager.number_of_colors = 8
module_manager2.scalar_lut_manager.use_default_name = False
module_manager2.scalar_lut_manager.scalar_bar.position = np.array([0.82, 0.1])
module_manager2.scalar_lut_manager.scalar_bar.position2 = np.array([0.17, 0.8])
module_manager2.scalar_lut_manager.scalar_bar.title = "Current density"
module_manager2.scalar_lut_manager.scalar_bar.position = np.array([0.82, 0.1])
module_manager2.scalar_lut_manager.scalar_bar.position2 = np.array([0.17, 0.8])
module_manager2.scalar_lut_manager.data_name = "Current density"
module_manager2.scalar_lut_manager.label_text_property.shadow_offset = np.array([1, -1])
module_manager2.scalar_lut_manager.label_text_property.italic = False
module_manager2.scalar_lut_manager.label_text_property.shadow_offset = np.array([1, -1])
module_manager2.scalar_lut_manager.label_text_property.color = (0.0, 0.0, 0.0)
module_manager2.scalar_lut_manager.title_text_property.shadow_offset = np.array([1, -1])
module_manager2.scalar_lut_manager.title_text_property.italic = False
module_manager2.scalar_lut_manager.title_text_property.shadow_offset = np.array([1, -1])
module_manager2.scalar_lut_manager.title_text_property.color = (0.0, 0.0, 0.0)
module_manager2.scalar_lut_manager.title_text_property.shadow_offset = np.array([1, -1])
module_manager2.scalar_lut_manager.title_text_property.bold = False

if SAVE:
    mlab.savefig(
        SAVE_DIR + "streamfunction_gradient.png",
        figure=f,
        magnification=4,
    )



import numpy as np
import matplotlib.pyplot as plt
import trimesh
from mayavi import mlab
from bfieldtools.thermal_noise import (
    compute_current_modes,
    noise_covar,
    noise_covar_dir,
    noise_var,
    visualize_current_modes,
)
from bfieldtools.mesh_magnetics import magnetic_field_coupling

import pkg_resources


font = {"family": "normal", "weight": "normal", "size": 16}

plt.rc("font", **font)

# Fix the simulation parameters
d = 1e-3
sigma = 3.8e7
T = 293
kB = 1.38064852e-23
mu0 = 4 * np.pi * 1e-7

Nchunks = 4
quad_degree = 2

freqs = np.array((0, 3000))


mesh = trimesh.load(
    pkg_resources.resource_filename("bfieldtools", "example_meshes/star_dense.stl")
)

vl = compute_current_modes(
    obj=mesh,
    T=T,
    resistivity=1 / sigma,
    thickness=d,
    mode="AC",
    freqs=freqs,
    return_eigenvals=False,
)


#%%
scene = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
Nmodes = 20
dist = 0.2

colormap = "bwr"

N1 = np.floor(np.sqrt(Nmodes))
dx = (mesh.vertices[:, 0].max() - mesh.vertices[:, 0].min()) * (1 + dist)
dy = (mesh.vertices[:, 1].max() - mesh.vertices[:, 1].min()) * (1 + dist)

i = 0
j = 0
for n in range(Nmodes):
    print(i, j)
    points = mesh.vertices.copy()
    points[:, 0] += i * dx
    points[:, 1] -= j * dy

    if n == 0:
        s = mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, color=(0, 0, 0))
        scene.scene.z_minus_view()
        surface = scene.children[0].children[0].children[0].children[0]
        surface.actor.property.representation = "wireframe"
    else:

        s = mlab.triangular_mesh(
            *points.T,
            mesh.faces,
            scalars=vl[:, n - 1, 0],
            colormap=colormap,
            line_width=6
        )

        limit = np.max(np.abs(vl[:, n]))

        s.module_manager.scalar_lut_manager.number_of_colors = 256
        s.module_manager.scalar_lut_manager.data_range = np.array([-limit, limit])
        s.actor.mapper.interpolate_scalars_before_mapping = True
        s.enable_contours = True
        s.contour.number_of_contours = 10

        mlab.triangular_mesh(*points.T, mesh.faces, color=(0, 0, 0), opacity=0.3)

    if i < N1:
        i += 1
    else:
        j += 1
        i = 0

# scene = engine.scenes[0]
scene.scene.z_plus_view()
scene.scene.camera.position = [
    2.4003626448678688,
    -1.7934540101222551,
    12.329167780858096,
]
scene.scene.camera.focal_point = [2.4003626448678688, -1.7934540101222551, 0.0]
scene.scene.camera.view_angle = 30.0
scene.scene.camera.view_up = [0.0, 1.0, 0.0]
scene.scene.camera.clipping_range = [11.548461583553994, 13.341654730403269]
scene.scene.camera.compute_view_plane_normal()
scene.scene.render()
scene.scene.camera.position = [
    2.4003626448678688,
    -1.7934540101222551,
    10.189394860213302,
]
scene.scene.camera.focal_point = [2.4003626448678688, -1.7934540101222551, 0.0]
scene.scene.camera.view_angle = 30.0
scene.scene.camera.view_up = [0.0, 1.0, 0.0]
scene.scene.camera.clipping_range = [9.544183126904127, 11.026160934217575]
scene.scene.camera.compute_view_plane_normal()
scene.scene.render()


#%%
Ngrid = 21
xx = np.linspace(np.min(mesh.vertices[:, 0]), np.max(mesh.vertices[:, 0]), Ngrid)
yy = np.linspace(np.min(mesh.vertices[:, 1]), np.max(mesh.vertices[:, 1]), Ngrid)
zz = np.max(mesh.vertices[:, 2]) + np.array([0.1, 0.2, 0.3, 0.4, 0.5])
X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

x = X.ravel()
y = Y.ravel()
z = Z.ravel()

fp = np.vstack((x, y, z)).T


B_coupling = magnetic_field_coupling(mesh, fp, analytic=True)

Bvar = noise_var(B_coupling, vl) * 1e30

#%%

cmap = "inferno"


vmin = 1e-2
vmax = 30


levels = np.linspace(vmin, vmax, 5)

Bzcov = np.sqrt(Bvar[:, 0, 0])
Bgrid = Bzcov.reshape((Ngrid, Ngrid, 5))


from matplotlib import colors
import matplotlib.gridspec as gridspec

Bzcov = np.sqrt(Bvar[:, 0, 0])
Bgrid = Bzcov.reshape((Ngrid, Ngrid, 5))

fig = plt.figure(figsize=(7, 2))
gs = gridspec.GridSpec(1, 5)
gs.update(wspace=0.025, hspace=0.025, left=0, right=0.9, top=1, bottom=0)


for ax_idx in range(5):
    ax = plt.subplot(gs[ax_idx])
    cont = ax.pcolormesh(
        X[:, :, ax_idx],
        Y[:, :, ax_idx],
        Bgrid[:, :, ax_idx],
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        norm=colors.LogNorm(),
        shading="gouraud",
    )
    ax.axis("equal")
    ax.axis("off")


cbar_ax = fig.add_axes([0.905, 0.2, 0.01, 0.6])
cbar = fig.colorbar(cont, cax=cbar_ax)


Bzcov = np.sqrt(Bvar[:, 1, 0])
Bgrid = Bzcov.reshape((Ngrid, Ngrid, 5))

fig = plt.figure(figsize=(7, 2))
gs = gridspec.GridSpec(1, 5)
gs.update(wspace=0.025, hspace=0.025, left=0, right=0.9, top=1, bottom=0)

for ax_idx in range(5):
    ax = plt.subplot(gs[ax_idx])
    cont = ax.pcolormesh(
        X[:, :, ax_idx],
        Y[:, :, ax_idx],
        Bgrid[:, :, ax_idx],
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        norm=colors.LogNorm(),
        shading="gouraud",
    )
    ax.axis("equal")
    ax.axis("off")


cbar_ax = fig.add_axes([0.905, 0.15, 0.01, 0.7])
cbar = fig.colorbar(cont, cax=cbar_ax)


Bzcov = np.sqrt(Bvar[:, 2, 0])
Bgrid = Bzcov.reshape((Ngrid, Ngrid, 5))

fig = plt.figure(figsize=(7, 2))
gs = gridspec.GridSpec(1, 5)
gs.update(wspace=0.025, hspace=0.025, left=0, right=0.9, top=1, bottom=0)

for ax_idx in range(5):
    ax = plt.subplot(gs[ax_idx])
    cont = ax.pcolormesh(
        X[:, :, ax_idx],
        Y[:, :, ax_idx],
        Bgrid[:, :, ax_idx],
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        norm=colors.LogNorm(),
        shading="gouraud",
    )
    ax.axis("equal")
    ax.axis("off")


cbar_ax = fig.add_axes([0.905, 0.2, 0.01, 0.6])
cbar = fig.colorbar(cont, cax=cbar_ax)


Bzcov = np.sqrt(Bvar[:, 2, 1])
Bgrid = Bzcov.reshape((Ngrid, Ngrid, 5))

fig = plt.figure(figsize=(7, 2))
gs = gridspec.GridSpec(1, 5)
gs.update(wspace=0.025, hspace=0.025, left=0, right=0.9, top=1, bottom=0)

for ax_idx in range(5):
    ax = plt.subplot(gs[ax_idx])
    cont = ax.pcolormesh(
        X[:, :, ax_idx],
        Y[:, :, ax_idx],
        Bgrid[:, :, ax_idx],
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        norm=colors.LogNorm(),
        shading="gouraud",
    )
    ax.axis("equal")
    ax.axis("off")


cbar_ax = fig.add_axes([0.905, 0.2, 0.01, 0.6])
cbar = fig.colorbar(cont, cax=cbar_ax)



"""
This module contains miscellaneous utility functions used across bfieldtools.
"""

__all__ = [
    "MeshProjection",
    "combine_meshes",
    "cylinder_points",
    "dual_areas",
    "find_mesh_boundaries",
    "fix_normals",
    "inner2vert",
    "load_example_mesh",
    "tri_normals_and_areas",
    "vert2inner",
]


import os
import numpy as np
import pkg_resources
import trimesh

from .quadratures import get_quad_points


def combine_meshes(meshes):
    """
    Combine two or more non-overlapping Trimesh meshes without any dependency
    requirements. For more demanding applications, use Trimesh boolean operations

    Parameters
    ----------
    meshes: list or tuple
        Each element should be a Trimesh mesh

    Returns
    -------
    combined_mesh: Trimesh mesh
    """

    N_meshes = len(meshes)

    vertices = np.zeros((0, 3))
    faces = np.zeros((0, 3))

    for idx in range(N_meshes):

        faces = np.vstack((faces, meshes[idx].faces + len(vertices)))
        vertices = np.vstack((vertices, meshes[idx].vertices))

    combined_mesh = trimesh.Trimesh(vertices=vertices, faces=faces, process=False)
    return combined_mesh


def tri_normals_and_areas(r, tri):
    """Get triangle normals and areas from vertices (r) and
    triangle indices (tri)
    """
    n = np.cross(
        r[tri[:, 1], :] - r[tri[:, 0], :], r[tri[:, 2], :] - r[tri[:, 0], :], axis=1
    )
    a = np.linalg.norm(n, axis=1)
    n = n / a[:, None]
    a /= 2
    return n, a


def dual_areas(tris, ta):
    """Calculate (dual) areas for each node in inds

    Dual area == area summed over the neighbouring triangles divided by 3
    """
    areas = np.zeros(np.max(tris) + 1)
    for i in range(tris.shape[0]):
        for j in range(tris.shape[1]):
            areas[tris[i, j]] += ta[i]

    return areas / 3


def find_mesh_boundaries(mesh):
    """
    Finds the open boundaries of a mesh by finding the edges that only
    belong to a single triangle. Returns an index array of inner vertices
    and triangles that do not touch the outer boundary.
    Takes edge parameter for convenience.

    Parameters
    ----------
    mesh: trimesh mesh object

    Returns
    -------
    boundaries: list of array-like

    """
    inner_vertices = np.arange(0, len(mesh.vertices))

    outline = mesh.outline(process=False)

    boundaries = []
    for i in outline.entities:
        boundaries.append(np.unique(i.points))

        inner_vertices = np.setdiff1d(inner_vertices, i.points)

    return boundaries, inner_vertices


def inner2vert(mesh, inner_vertices, holes):
    """Linear mapping of the inner (free) weights in the stream function
    discretization to weights in all vertices

    Parameters:
        mesh: Trimesh object
        inner_vertices: list of indices of the inner vertices of the mesh
        holes: list of indices for holes in the mesh

    Returns:
        NxM sparse array, where N==mesh.vertices.shape[0]
        and M == len(inner_vertices) + len(holes)
    """
    from scipy.sparse import csr_matrix

    N = mesh.vertices.shape[0]
    M = len(inner_vertices) + len(holes)
    ii = list(inner_vertices)  # indices of inner vertices
    jj = list(np.arange(len(inner_vertices)))  # indices of inner vertices in dof
    # Hole values maps to value for each hole vertex
    for n, h in enumerate(holes):
        ii.extend(list(h))  # indices of hole indices
        jj.extend(
            [len(inner_vertices) + n] * len(h)
        )  # len(h) times index of hole in dof
    d2v = csr_matrix((np.ones(len(jj)), (ii, jj)), shape=(N, M), dtype=float)

    return d2v


def vert2inner(mesh, inner_vertices, holes):
    """Linear mapping of the all weights in the stream function
    discretization to inner (free) weights

    Parameters:
        mesh: Trimesh object
        inner_vertices: list of indices of the inner vertices of the mesh
        holes: list of indices for holes in the mesh

    Returns:
        MxN sparse array, where N==mesh.vertices.shape[0]
        and M == len(inner_vertices) + len(holes)
    """
    from scipy.sparse import csr_matrix

    N = mesh.vertices.shape[0]
    M = len(inner_vertices) + len(holes)
    ii = list(inner_vertices)  # indices of inner vertices
    jj = list(np.arange(len(inner_vertices)))  # indices of inner vertices in dof
    vals = list(np.ones(len(inner_vertices)))
    for n, h in enumerate(holes):
        ii.extend(list(h))  # indices of hole indices
        jj.extend(
            [len(inner_vertices) + n] * len(h)
        )  # len(h) times index of hole in free values
        # Values at holes map to their average (ok, when constant boundary condition satisfied)
        vals.extend(list(np.ones(len(h)) / len(h)))
    v2d = csr_matrix((vals, (jj, ii)), shape=(M, N), dtype=float)

    return v2d


def cylinder_points(
    radius=1,
    length=1,
    nlength=10,
    alpha=360,
    nalpha=10,
    center=np.array([0, 0, 0]),
    orientation=np.array([1, 0, 0]),
):
    """
    Generate and return a set of points on a cylindrical surface.
    """
    # Create the length array
    I = np.linspace(0, length, nlength)

    # Create alpha array avoid duplication of endpoints
    # Conditional should be changed to meet your requirements
    if int(alpha) == 360:
        A = np.linspace(0, alpha, num=nalpha, endpoint=False) / 180 * np.pi
    else:
        A = np.linspace(0, alpha, num=nalpha) / 180 * np.pi

    # Calculate X and Y
    X = radius * np.cos(A)
    Y = radius * np.sin(A)

    # Tile/repeat indices so all unique pairs are present
    pz = np.tile(I, nalpha)
    px = np.repeat(X, nlength)
    py = np.repeat(Y, nlength)

    points = np.vstack((pz, px, py)).T

    # Shift to center
    shift = np.array(center) - np.mean(points, axis=0)
    points += shift

    # Orient tube to new vector

    def rotation_matrix(axis, theta):
        a = np.cos(theta / 2)
        b, c, d = -axis * np.sin(theta / 2)
        return np.array(
            [
                [
                    a * a + b * b - c * c - d * d,
                    2 * (b * c - a * d),
                    2 * (b * d + a * c),
                ],
                [
                    2 * (b * c + a * d),
                    a * a + c * c - b * b - d * d,
                    2 * (c * d - a * b),
                ],
                [
                    2 * (b * d - a * c),
                    2 * (c * d + a * b),
                    a * a + d * d - b * b - c * c,
                ],
            ]
        )

    ovec = orientation / np.linalg.norm(orientation)
    cylvec = np.array([1, 0, 0])

    if np.allclose(cylvec, ovec):
        return points

    # Get orthogonal axis and rotation
    oaxis = np.cross(ovec, cylvec)
    rot = np.arccos(np.dot(ovec, cylvec))

    R = rotation_matrix(oaxis, rot)
    return points.dot(R)


def fix_normals(mesh, origin=np.array([0, 0, 0])):
    """
    Attempts to fix face windings and normals such that normals are always "pointing out"
    from the origin.

    Parameters
    ----------
    mesh: Trimesh mesh object
    origin: array-like (3, )
        Specified from where the normals should "point out"

    Returns
    -------
    mesh: modified Trimesh object


    """

    # Dot product of all face normals and the corresponding triangle centers
    dotprods = np.sum(mesh.face_normals * (mesh.triangles_center - origin), axis=-1)
    # Flip windings for normals pointing inwards
    mesh.faces[dotprods < 0] = mesh.faces[dotprods < 0, ::-1]
    # old_cache = {'key': mesh.cache.value, ...} # Could save some old values...
    # Clear cached values
    mesh._cache.clear()
    # Could update the cache with values that don't change when flipping triangles
    # self._cache.update(old_cache)
    return mesh


def load_example_mesh(mesh_name, process=True, **kwargs):
    """
    Convenience function used load example meshes included with the package

    Parameters
    ----------
    mesh_name: string
        name of mesh, i.e. filename without extension
    process: Boolean
        Whether trimesh should process the mesh on loading
    kwargs
        Passed to trimesh object creation

    Returns
    -------
    Trimesh object
    """
    existing_files = pkg_resources.resource_listdir("bfieldtools", "example_meshes")

    # Filter according to file extension
    existing_files = [
        file
        for file in existing_files
        if file.lower().endswith(tuple(trimesh.exchange.load.mesh_formats()))
    ]

    # Remove file extension to get name
    existing_names = [os.path.splitext(file)[0] for file in existing_files]

    # Check if name exists
    if mesh_name not in existing_names:
        raise ValueError(
            "Mesh with name %s not found in example_meshes folder" % mesh_name
        )

    filename = existing_files[existing_names.index(mesh_name)]

    return trimesh.load(
        pkg_resources.resource_filename("bfieldtools", "example_meshes/" + filename),
        process=process,
        **kwargs
    )


class MeshProjection:
    """Class for constructing projection of an arbitrary function
    to the hat functions of the given mesh

    The constructor initializes quadrature points, point weights
    and vertex mappings

    After the construction the method "hatfunc_innerproducts" can
    be called to obtain inner products of the hat functions and
    a given function (func)
    """

    def __init__(self, mesh, quad_degree):
        self.mesh = mesh

        weights, self.quadpoints, ref_coords = get_quad_points(
            mesh.vertices, mesh.faces, ("dunavant", quad_degree), return_ref_coords=True
        )
        # Mapping from coordinates of a reference triangle to barycentric coordinates
        self.M = np.array([[-1.0, -1.0, 1.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]])
        # Solve barycentric coordinates for the quad points
        # These are the shape function values for the quad points
        self.bary_coords = (
            np.hstack((ref_coords, np.ones((ref_coords.shape[0], 1)))) @ self.M.T
        )
        # Weight the shape function values by quadrature weights
        self.weights = self.bary_coords * weights[:, None]
        from scipy.sparse import csc_matrix

        Nf = len(mesh.faces)
        Nv = len(mesh.vertices)
        self.M0 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 0])), (Nf, Nv))
        self.M1 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 1])), (Nf, Nv))
        self.M2 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 2])), (Nf, Nv))

    def hatfunc_innerproducts(self, func):
        ff = np.zeros((self.quadpoints.shape[0], 3))
        for i in range(self.quadpoints.shape[1]):
            f = func(self.quadpoints[:, i, :])
            ff += self.weights[i] * f[:, None]

        ff *= self.mesh.area_faces[:, None]
        fv = ff[:, 0] @ self.M0 + ff[:, 1] @ self.M1 + ff[:, 2] @ self.M2

        return fv



from ..mesh_impedance import mutual_inductance_matrix

import numpy as np


def self_inductance_matrix(
    mesh,
    Nchunks=None,
    quad_degree=2,
    approx_far=True,
    margin=2,
    chunk_clusters=False,
):
    """Calculate a self inductance matrix for hat basis functions
    (stream functions) in the triangular mesh described by

    Parameters
    ----------
    mesh: Trimesh mesh object
    Nchunks: int
        Number of serial chunks to divide the computation into
    quad_degree: int >= 1
        Quadrature degree (Dunavant scheme) to use. Self-inductance requires higher degree than mutual inductance
    approx_far: Boolean (True)
        If True, use approximate calculation for triangles that
        far from the source triangles using a simple quadrature
        (see integrals.triangle_potential_approx)
    margin: float
        Cut-off distance for "far" points measured in mean triangle side length

    Returns
    -------
    M: (Nvertices x Nvertices) array
        Self.inductance matrix of `mesh`
    """
    if quad_degree <= 2:
        print(
            "Computing self-inductance matrix using rough quadrature (degree=%d).\
              For higher accuracy, set quad_degree to 4 or more."
            % quad_degree
        )

    return mutual_inductance_matrix(
        mesh,
        mesh,
        Nchunks=Nchunks,
        quad_degree=quad_degree,
        approx_far=approx_far,
        margin=margin,
    )


def triangle_self_coupling(mesh):
    """
    Self-coupling integrated analytically. Implemented based on
    https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=475946

    Parameters
    ----------
    mesh: Trimesh mesh object

    Returns
    -------
    self_coupling: array (N_triangles, )
        triangle self-coupling
    """

    tri_points = mesh.vertices[mesh.faces]

    r1 = tri_points[:, 0, :]
    r2 = tri_points[:, 1, :]
    r3 = tri_points[:, 2, :]

    a = np.einsum("ij,ij->i", r3 - r1, r3 - r1)
    b = np.einsum("ij,ij->i", r3 - r1, r3 - r2)
    c = np.einsum("ij,ij->i", r3 - r2, r3 - r2)

    sa = np.sqrt(a)
    sc = np.sqrt(c)
    ss = np.sqrt(a - 2 * b + c)
    sac = np.sqrt(a * c)

    self_coupling = (1 * (4 * mesh.area_faces**2)) * (
        1
        / (6 * sa)
        * np.log(((a - b + sa * ss) * (b + sac)) / ((-b + sac) * (-a + b + sa * ss)))
        + 1
        / (6 * sc)
        * np.log(((b + sac) * (-b + c + sc * ss)) / ((b - c + sc * ss) * (-b + sac)))
        + 1
        / (6 * ss)
        * np.log(
            ((a - b + sa * ss) * (-b + c + sc * ss))
            / ((b - c + sc * ss) * (-a + b + sa * ss))
        )
    )

    return self_coupling
